# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway
# Copyright 2021, Gaute Hope, MET Norway
"""
OpenOil is a 3D oil drift module bundled within the OpenDrift framework.

The oil weathering calculations is based on the NOAA-ERR-ERD OilLibrary
package, which is installed as a dependency. The code for evaporation and
emulsification in OpenOil is borrowed from the NOAA PyGnome code, and adapted
to the OpenDrift architecture.

Example of ship leaking oil along the coast of Northern Norway
##############################################################

.. image:: https://dl.dropboxusercontent.com/s/ty6dmqf0oohewky/oilspill_tromsoe.gif?dl=0

Simulation of Deepwater Horizon (Macondo) accident, initiated from satellite images
###################################################################################

.. image:: https://dl.dropboxusercontent.com/s/ghi7crtmwpyjgto/macondo_simulation.gif?dl=0

Satellite images provided by Prof. Chuanmin Hu, and ocean model output provided by Prof. Vassiliki Kourafalou

Example oil budget for a simulation
###################################

.. image:: https://dl.dropboxusercontent.com/s/pb0h6tlev9pnoh3/oil_budget_draugen.png?dl=0

Oil properties affecting the drift
***********************************
The vertical (and thus indirectly also the horizontal) motion of oil (droplets) is affected by oil density and droplet diameters.

When using the NOAA oil weathering model (``o = OpenOil(weathering_model='noaa')``), which is the default, the density is obtained from the NOAA database according to the oiltype selected when seeding. This value can not be overridden by the user, and it will also change during the simulation due to oil weathering processes (evaporation and emulsification).

The droplet diameter may be given explicitly when seeding, e.g.:

.. code::

    o.seed_elements(4, 60, number=100, time=datetime.now(), diameter=1e-5)

In this case, the diameter will not change during the simulation, which is useful e.g. for sensitivity tests. The same diameter will be used for all elements for this example, but an array of the same length as the number of elements may also be provided.

If a constant droplet diameter is not given by the user, it will be chosen randomly within given config limits for a subsea spill ('blowout'), and modified after any later wave breaking event. Oil droplets seeded under sea surface (z<0) will be assigned initial diameters between the following limits, typical for a subsea blowout (Johansen, 2000)::

.. code::

    o.set_config('seed:droplet_diameter_min_subsea', 0.0005)  # 0.5 mm
    o.set_config('seed:droplet_diameter_max_subsea', 0.005)   # 5 mm

Note that these config settings must be adjusted before the seeding call.
After each wave breaking event, a new droplet diameter will be chosen based on the config setting for droplet size distribution.
"""

from io import open
import numpy as np
from datetime import datetime
import pyproj
import matplotlib.pyplot as plt
import logging

logger = logging.getLogger(__name__)

from opendrift.models.oceandrift import OceanDrift, Lagrangian3DArray
from . import noaa_oil_weathering as noaa
from . import adios
from opendrift.models.physics_methods import oil_wave_entrainment_rate_li2017


# Defining the oil element properties
class Oil(Lagrangian3DArray):
    """Extending LagrangianArray with variables relevant for oil particles."""

    variables = Lagrangian3DArray.add_variables([
        ('mass_oil', {
            'dtype': np.float32,
            'units': 'kg',
            'seed': False,
            'default': 1
        }),
        (
            'viscosity',
            {
                'dtype': np.float32,
                'units': 'N s/m2 (Pa s)',
                'seed': False,  # Taken from NOAA database
                'default': 0.005
            }),
        (
            'density',
            {
                'dtype': np.float32,
                'units': 'kg/m^3',
                'seed': False,  # Taken from NOAA database
                'default': 880
            }),
        (
            'wind_drift_factor',
            {
                'dtype':
                np.float32,  # TODO: inherit from
                'units':
                '%',  # OceanDrift
                'description':
                'Elements at the ocean surface are moved by '
                'this fraction of the wind vector, in addition to '
                'currents and Stokes drift',
                'default':
                0.03
            }),
        ('bulltime', {
            'dtype': np.float32,
            'units': 's',
            'seed': False,
            'default': 0
        }),
        ('interfacial_area', {
            'dtype': np.float32,
            'units': 'm2',
            'seed': False,
            'default': 0
        }),
        ('mass_dispersed', {
            'dtype': np.float32,
            'units': 'kg',
            'seed': False,
            'default': 0
        }),
        ('mass_evaporated', {
            'dtype': np.float32,
            'units': 'kg',
            'seed': False,
            'default': 0
        }),
        ('mass_biodegraded', {
            'dtype': np.float32,
            'units': 'kg',
            'seed': False,
            'default': 0
        }),
        ('fraction_evaporated', {
            'dtype': np.float32,
            'units': '%',
            'seed': False,
            'default': 0
        }),
        ('water_fraction', {
            'dtype': np.float32,
            'units': '%',
            'seed': False,
            'default': 0
        }),
        ('oil_film_thickness', {
            'dtype': np.float32,
            'units': 'm',
            'default': 0.001
        }),
        (
            'diameter',
            {
                'dtype': np.float32,  # Particle diameter
                'units': 'm',
                'seed': False,
                'default': 0.
            })
    ])


class OpenOil(OceanDrift):
    """Open source oil trajectory model based on the OpenDrift framework.

        Developed at MET Norway based on oil weathering parameterisations
        found in open/published litterature.

        Under construction.
    """

    ElementType = Oil

    required_variables = {
        'x_sea_water_velocity': {
            'fallback': None
        },
        'y_sea_water_velocity': {
            'fallback': None
        },
        'x_wind': {
            'fallback': None
        },
        'y_wind': {
            'fallback': None
        },
        'upward_sea_water_velocity': {
            'fallback': 0,
            'important': False
        },
        'sea_surface_wave_significant_height': {
            'fallback': 0,
            'important': False
        },
        'sea_surface_wave_stokes_drift_x_velocity': {
            'fallback': 0,
            'important': False
        },
        'sea_surface_wave_stokes_drift_y_velocity': {
            'fallback': 0,
            'important': False
        },
        'sea_surface_wave_period_at_variance_spectral_density_maximum': {
            'fallback': 0,
            'important': False
        },
        'sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment':
        {
            'fallback': 0,
            'important': False
        },
        'sea_ice_area_fraction': {
            'fallback': 0,
            'important': False
        },
        'sea_ice_x_velocity': {
            'fallback': 0,
            'important': False
        },
        'sea_ice_y_velocity': {
            'fallback': 0,
            'important': False
        },
        'sea_water_temperature': {
            'fallback': 10,
            'profiles': True
        },
        'sea_water_salinity': {
            'fallback': 34,
            'profiles': True
        },
        'sea_floor_depth_below_sea_level': {
            'fallback': 10000
        },
        'ocean_vertical_diffusivity': {
            'fallback': 0.02,
            'important': False,
            'profiles': True
        },
        'land_binary_mask': {
            'fallback': None
        },
        'ocean_mixed_layer_thickness': {
            'fallback': 50,
            'important': False
        },
    }

    # The depth range (in m) which profiles shall cover
    required_profiles_z_range = [-20, 0]

    max_speed = 1.3  # m/s

    # Default colors for plotting
    status_colors = {
        'initial': 'green',
        'active': 'blue',
        'missing_data': 'gray',
        'stranded': 'red',
        'evaporated': 'yellow',
        'dispersed': 'magenta'
    }

    duplicate_oils = ['ALVHEIM BLEND, STATOIL', 'DRAUGEN, STATOIL',
                  'EKOFISK BLEND 2000', 'EKOFISK BLEND, STATOIL',
                  'EKOFISK, CITGO', 'EKOFISK, EXXON', 'EKOFISK, PHILLIPS',
                  'EKOFISK, STATOIL', 'ELDFISK', 'ELDFISK B',
                  'GLITNE, STATOIL', 'GOLIAT BLEND, STATOIL',
                  'GRANE BLEND, STATOIL', 'GUDRUN BLEND, STATOIL',
                  'GULLFAKS A, STATOIL', 'GULLFAKS C, STATOIL',
                  'GULLFAKS, SHELL OIL', 'GULLFAKS SOR',
                  'GULLFAKS, STATOIL', 'HEIDRUN, STATOIL',
                  'NJORD, STATOIL', 'NORNE, STATOIL',
                  'OSEBERG BLEND, STATOIL', 'OSEBERG EXXON',
                  'OSEBERG, PHILLIPS', 'OSEBERG, SHELL OIL',
                  'SLEIPNER CONDENSATE, STATOIL',
                  'STATFJORD BLEND, STATOIL', 'VARG, STATOIL']

    # Workaround as ADIOS oil library uses
    # max water fraction of 0.9 for all crude oils
    max_water_fraction = {
        'MARINE GAS OIL 500 ppm S 2017': 0.1,
        'FENJA (PIL) 2015': .75
    }

    def __init__(self, weathering_model='noaa', *args, **kwargs):
        self.oil_weathering_model = weathering_model

        if self.oil_weathering_model == 'noaa':  # Currently the only option
            self.oiltypes = adios.get_oil_names(
                location=kwargs.get('location', None))

            # Update config with oiltypes
            self.oiltypes.extend(adios.oil_name_alias.keys())

            # Sort alphabetically, but put GENERIC oils first
            generic_oiltypes = [o for o in self.oiltypes if o[0:7] == 'GENERIC']
            other_oiltypes = [o for o in self.oiltypes if o[0:7] != 'GENERIC']
            self.oiltypes = sorted([o for o in generic_oiltypes]) + sorted([o for o in other_oiltypes])
            self.oiltypes = [ot for ot in self.oiltypes if ot not in self.duplicate_oils]
        else:
            raise ValueError('Weathering model unknown: ' + weathering_model)

        kwargs.pop('location', None)

        # Calling general constructor of parent class
        super(OpenOil, self).__init__(*args, **kwargs)

        self._add_config({
            'seed:m3_per_hour': {
                'type': 'float',
                'default': 1,
                'min': 0,
                'max': 1e10,
                'units': 'm3 per hour',
                'description':
                'The amount (volume) of oil released per hour (or total amount if release is instantaneous)',
                'level': self.CONFIG_LEVEL_ESSENTIAL
            },
            'seed:droplet_diameter_min_subsea': {
                'type': 'float',
                'default': 0.0005,
                'min': 1e-8,
                'max': 1,
                'units': 'meters',
                'description':
                'The minimum diameter of oil droplet for a subsea release.',
                'level': self.CONFIG_LEVEL_BASIC
            },
            'seed:droplet_diameter_max_subsea': {
                'type': 'float',
                'default': 0.005,
                'min': 1e-8,
                'max': 1,
                'units': 'meters',
                'description':
                'The maximum diameter of oil droplet for a subsea release.',
                'level': self.CONFIG_LEVEL_BASIC
            },
            'processes:dispersion': {
                'type': 'bool',
                'default': True,
                'description':
                'Oil is removed from simulation (dispersed), if entrained as very small droplets.',
                'level': self.CONFIG_LEVEL_BASIC
            },
            'processes:evaporation': {
                'type': 'bool',
                'default': True,
                'description': 'Surface oil is evaporated.',
                'level': self.CONFIG_LEVEL_BASIC
            },
            'processes:emulsification': {
                'type': 'bool',
                'default': True,
                'description':
                'Surface oil is emulsified, i.e. water droplets are mixed into oil due to wave mixing, with resulting increas of viscosity.',
                'level': self.CONFIG_LEVEL_BASIC
            },
            'processes:biodegradation': {
                'type': 'bool',
                'default': False,
                'description': 'Oil mass is biodegraded (eaten by bacteria).',
                'level': self.CONFIG_LEVEL_BASIC
            },
            'processes:update_oilfilm_thickness': {
                'type': 'bool',
                'default': False,
                'description':
                'Oil film thickness is calculated at each time step. The alternative is that oil film thickness is kept constant with value provided at seeding.',
                'level': self.CONFIG_LEVEL_ADVANCED
            },
            'wave_entrainment:droplet_size_distribution': {
                'type':
                'enum',
                'enum': ['Johansen et al. (2015)', 'Li et al. (2017)'],
                'default':
                'Johansen et al. (2015)',
                'level':
                self.CONFIG_LEVEL_ADVANCED,
                'description':
                'Algorithm to be used for calculating oil droplet size spectrum after entrainment by breaking waves.'
            },
            'wave_entrainment:entrainment_rate': {
                'type':
                'enum',
                'enum': ['Li et al. (2017)'],
                'default':
                'Li et al. (2017)',
                'level':
                self.CONFIG_LEVEL_ADVANCED,
                'description':
                'Algorithm to be used for calculating the entrainment rate of oil due to wave breaking.'
            },
            'seed:oil_type': {
                'type':
                'enum',
                'enum':
                self.oiltypes,
                'default':
                self.oiltypes[0],
                'level':
                self.CONFIG_LEVEL_ESSENTIAL,
                'description':
                'Oil type to be used for the simulation, from the NOAA ADIOS database.'
            },
        })

        self._set_config_default('drift:vertical_advection', False)
        self._set_config_default('drift:vertical_mixing', True)
        self._set_config_default('drift:current_uncertainty', 0.05)
        self._set_config_default('drift:wind_uncertainty', 0.5)

    def update_surface_oilfilm_thickness(self):
        '''The mass of oil is summed within a grid of 100x100
        cells covering the oil at a given time. Each oil particle
        within each cell is given a film thickness as the amount of
        oil divided by the cell area.
        '''
        from scipy.stats import binned_statistic_2d
        surface = np.where(self.elements.z == 0)[0]
        if len(surface) == 0:
            logger.debug('No oil at surface, no film thickness to update')
            return
        logger.debug(
            'Updating oil film thickness for %s of %s elements at surface' %
            (len(surface), self.num_elements_active()))
        meanlon = self.elements.lon[surface].mean()
        meanlat = self.elements.lat[surface].mean()
        # Using stereographic coordinates to get regular X and Y
        psproj = pyproj.Proj('+proj=stere +lat_0=%s +lat_ts=%s +lon_0=%s' %
                             (meanlat, meanlat, meanlon))
        X, Y = psproj(self.elements.lon[surface], self.elements.lat[surface])
        mass_bin, x_edge, y_edge, binnumber = binned_statistic_2d(
            X,
            Y,
            self.elements.mass_oil[surface],
            expand_binnumbers=True,
            statistic='sum',
            bins=100)
        bin_area = (x_edge[1] - x_edge[0]) * (y_edge[1] - y_edge[0])
        oil_density = 1000  # ok approximation here
        film_thickness = (mass_bin / oil_density) / bin_area
        # Postulating min and max film thickness
        max_thickness = 0.01  # 1 cm
        min_thickness = 1e-9  # 1 nanometer
        if film_thickness.max() > max_thickness:
            logger.debug(
                'Warning: decreasing thickness to %sm for %s of %s bins' %
                (max_thickness, np.sum(film_thickness > max_thickness),
                 film_thickness.size))
            film_thickness[film_thickness > max_thickness] = max_thickness
        num_too_thin = np.sum((film_thickness < min_thickness)
                              & (film_thickness > 0))
        if num_too_thin > 0:
            logger.debug(
                'Warning: increasing thickness to %sm for %s of %s bins' %
                (min_thickness, num_too_thin, film_thickness.size))
            film_thickness[film_thickness < min_thickness] = min_thickness

        # https://github.com/scipy/scipy/issues/7010
        binnumber = binnumber - 1

        bx = binnumber[0, :]
        by = binnumber[1, :]
        # Update thickness
        self.elements.oil_film_thickness[
            surface] = self.elements.oil_film_thickness[surface] * np.nan
        self.elements.oil_film_thickness[surface] = \
            film_thickness[bx, by]

    def biodegradation(self):
        if self.get_config('processes:biodegradation') is True:
            '''
            Oil biodegradation function based on the article:
            Adcroft et al. (2010), Simulations of underwater plumes of
            dissolved oil in the Gulf of Mexico.
            '''
            logger.debug('Calculating: biodegradation')

            swt = self.environment.sea_water_temperature.copy()
            swt[swt > 100] -= 273.15  # K to C
            age0 = self.time_step.total_seconds() / (3600 * 24)

            # Decay rate in days (temperature in Celsius)
            tau = (12) * (3**((20 - swt) / 10))

            fraction_biodegraded = (1 - np.exp(-age0 / tau))
            biodegraded_now = self.elements.mass_oil * fraction_biodegraded

            self.elements.mass_biodegraded = \
                self.elements.mass_biodegraded + biodegraded_now
            self.elements.mass_oil = \
                self.elements.mass_oil - biodegraded_now
            if self.oil_weathering_model == 'noaa':
                self.noaa_mass_balance['mass_components'][self.elements.ID - 1, :] = \
                self.noaa_mass_balance['mass_components'][self.elements.ID - 1, :]*(1-fraction_biodegraded[:, np.newaxis])
            else:
                pass

    def disperse(self):
        if self.get_config('processes:dispersion') is True:

            logger.debug('   Calculating: dispersion')
            windspeed = np.sqrt(self.environment.x_wind**2 +
                                self.environment.y_wind**2)
            # From NOAA PyGnome model:
            # https://github.com/NOAA-ORR-ERD/PyGnome/
            v_entrain = 3.9E-8
            sea_water_density = 1028
            fraction_breaking_waves = 0.02
            wave_significant_height = \
                self.significant_wave_height()
            wave_significant_height[wave_significant_height == 0] = \
                0.0246*windspeed[wave_significant_height == 0]**2
            dissipation_wave_energy = \
                (0.0034*sea_water_density*9.81*wave_significant_height**2)
            c_disp = np.power(dissipation_wave_energy, 0.57) * \
                fraction_breaking_waves
            # Roy's constant
            C_Roy = 2400.0 * np.exp(-73.682 * np.sqrt(
                self.elements.viscosity / self.elements.density))

            q_disp = C_Roy * c_disp * v_entrain / self.elements.density

            oil_mass_loss = (q_disp * self.time_step.total_seconds() *
                             self.elements.density) * self.elements.mass_oil

            self.elements.mass_oil -= oil_mass_loss
            self.elements.mass_dispersed += oil_mass_loss

            #self.elements.z = \
            #    self.environment.sea_surface_wave_significant_height

            ## Marks R. (1987), Marine aerosols and whitecaps in the
            ## North Atlantic and Greenland sea regions
            ## Deutsche Hydrografische Zeitschrift, Vol 40, Issue 2 , pp 71-79
            #whitecap_coverage = (2.54E-4)*np.power(windspeed, 3.58)

            ## Martinsen et al. (1994), The operational
            ## oil drift system at DNMI
            ## DNMI Technical report No 125, 51
            #wave_period = 3.85*np.sqrt(
            #    self.environment.sea_surface_wave_significant_height)  # sec

            #time_between_breaking_events = 3.85/whitecap_coverage  # TBC

            #rho_w = 1025  # kg/m3
            #wave_period[wave_period == 0] = 5  # NB: temporal fix for no waves
            #dissipation_energy = 0.0034*rho_w*9.81*(wave_period**2)
            #dsize_coeff = 2100  # Wettre et al., Appendix A
            #c_oil = dsize_coeff*np.power(self.elements.viscosity, -0.4)
            ## Random numbers between 0 and 1:
            #p = np.random.rand(self.num_elements_active(), )
            #oil_per_unit_surface = 1
            #droplet_size = np.power((p*oil_per_unit_surface)/(c_oil *
            #                        np.power(dissipation_energy, 0.57)),
            #                        1/1.17)
            #self.deactivate_elements(droplet_size < 5E-7, reason='dispersed')

    def oil_weathering(self):
        if self.time_step.days < 0:
            logger.debug('Skipping oil weathering for backwards run')
            return
        self.timer_start('main loop:updating elements:oil weathering')
        if self.oil_weathering_model == 'noaa':
            self.oil_weathering_noaa()
        self.timer_end('main loop:updating elements:oil weathering')

    def prepare_run(self):

        if self.oil_weathering_model == 'noaa':
            self.noaa_mass_balance = {}
            # Populate with seeded mass spread on oiltype.mass_fraction
            mass_oil = np.atleast_1d(self.elements_scheduled.mass_oil)
            if len(mass_oil) == 1:
                mass_oil = mass_oil * np.ones(self.num_elements_total())
            self.noaa_mass_balance['mass_components'] = \
                np.asarray(self.oiltype.mass_fraction)*(mass_oil.reshape(
                    (self.num_elements_total(), -1)))
            self.noaa_mass_balance['mass_evaporated'] = \
                self.noaa_mass_balance['mass_components']*0

            self.oil_water_interfacial_tension = \
                self.oiltype.oil_water_surface_tension()
            logger.info('Oil-water surface tension is %f Nm' %
                        self.oil_water_interfacial_tension)

    def oil_weathering_noaa(self):
        '''Oil weathering scheme adopted from NOAA PyGNOME model:
        https://github.com/NOAA-ORR-ERD/PyGnome
        '''
        logger.debug('NOAA oil weathering')
        # C to K
        self.environment.sea_water_temperature[
            self.environment.sea_water_temperature < 100] += 273.15

        #########################################################
        # Update density and viscosity according to temperature
        #########################################################
        self.timer_start(
            'main loop:updating elements:oil weathering:updating viscosities')
        oil_viscosity = self.oiltype.kvis_at_temp(
            self.environment.sea_water_temperature)
        self.timer_end(
            'main loop:updating elements:oil weathering:updating viscosities')
        self.timer_start(
            'main loop:updating elements:oil weathering:updating densities')
        oil_density = self.oiltype.density_at_temp(
            self.environment.sea_water_temperature)
        self.timer_end(
            'main loop:updating elements:oil weathering:updating densities')

        # Calculate emulsion density
        self.elements.density = (
            self.elements.water_fraction * self.sea_water_density() +
            (1 - self.elements.water_fraction) * oil_density)

        # Calculate emulsion viscosity
        visc_f_ref = 0.84  # From PyGNOME
        visc_curvfit_param = 1.5e3  # units are sec^0.5 / m
        fw_d_fref = self.elements.water_fraction / visc_f_ref
        kv1 = np.sqrt(oil_viscosity) * visc_curvfit_param
        kv1[kv1 < 1] = 1
        kv1[kv1 > 10] = 10
        self.elements.fraction_evaporated = self.elements.mass_evaporated / (
            self.elements.mass_oil + self.elements.mass_evaporated)
        self.elements.viscosity = (
            oil_viscosity * np.exp(kv1 * self.elements.fraction_evaporated) *
            (1 + (fw_d_fref / (1.187 - fw_d_fref)))**2.49)

        if self.get_config('processes:evaporation') is True:
            self.timer_start(
                'main loop:updating elements:oil weathering:evaporation')
            self.evaporation_noaa()
            self.timer_end(
                'main loop:updating elements:oil weathering:evaporation')

        if self.get_config('processes:emulsification') is True:
            self.timer_start(
                'main loop:updating elements:oil weathering:emulsification')
            self.emulsification_noaa()
            self.timer_end(
                'main loop:updating elements:oil weathering:emulsification')

        if self.get_config('processes:dispersion') is True:
            self.timer_start(
                'main loop:updating elements:oil weathering:dispersion')
            self.disperse_noaa()
            self.timer_end(
                'main loop:updating elements:oil weathering:dispersion')

        if self.get_config('processes:biodegradation') is True:
            self.timer_start(
                'main loop:updating elements:oil weathering:biodegradation')
            self.biodegradation()
            self.timer_end(
                'main loop:updating elements:oil weathering:biodegradation')

        if self.elements.mass_oil.min() < 0:  # Should not happen
            logger.warning('NEGATIVE OIL MASS!')

    def disperse_noaa(self):
        logger.debug('    Calculating: dispersion - NOAA')
        # From NOAA PyGnome model:
        # https://github.com/NOAA-ORR-ERD/PyGnome/
        c_disp = np.power(self.wave_energy_dissipation(), 0.57) * \
            self.sea_surface_wave_breaking_fraction()
        # Roy's constant
        C_Roy = 2400.0 * np.exp(
            -73.682 * np.sqrt(self.elements.viscosity / self.elements.density))
        v_entrain = 3.9E-8
        q_disp = C_Roy * c_disp * v_entrain / self.elements.density
        fraction_dispersed = (q_disp * self.time_step.total_seconds() *
                              self.elements.density)
        if fraction_dispersed.max() >= 1:
            logger.warning(
                'Dispersion fraction larger than 1 -> setting to 0.99')
            fraction_dispersed[fraction_dispersed >= 1] = .99
        oil_mass_loss = fraction_dispersed * self.elements.mass_oil

        self.noaa_mass_balance['mass_components'][self.elements.ID - 1, :] = \
            self.noaa_mass_balance['mass_components'][self.elements.ID - 1, :]*(1-fraction_dispersed[:, np.newaxis])

        self.elements.mass_oil -= oil_mass_loss
        self.elements.mass_dispersed += oil_mass_loss

    def plot_droplet_spectrum(self):
        '''Plotting distribution of droplet radii, for debugging'''
        plt.hist(self.elements.diameter / 2.0)
        plt.show()

    def evaporation_noaa(self):
        #############################################
        # Evaporation, for elements at surface only
        #############################################
        logger.debug('    Calculating evaporation - NOAA')
        surface = np.where(self.elements.z == 0)[0]  # of active elements
        if len(surface) == 0:
            logger.debug('All elements submerged, no evaporation')
            return
        if self.elements.age_seconds[surface].min() > 3600 * 24:
            logger.debug('All surface oil elements older than 24 hours, ' +
                         'skipping further evaporation.')
            return
        surfaceID = self.elements.ID[surface] - 1  # of any elements
        # Area for each element, repeated for each component
        volume = (self.elements.mass_oil[surface] /
                  self.elements.density[surface])
        area = volume / self.elements.oil_film_thickness[surface]
        evap_decay_constant = noaa.evap_decay_constant(
            self.oiltype,
            self.wind_speed()[surface],
            self.environment.sea_water_temperature[surface], area,
            self.noaa_mass_balance['mass_components'][surfaceID, :])
        mass_remain = \
            (self.noaa_mass_balance['mass_components'][surfaceID, :] *
             np.exp(evap_decay_constant*self.time_step.total_seconds()))
        self.elements.mass_evaporated[surface] += \
            np.sum(self.noaa_mass_balance[
                    'mass_components'][surfaceID, :] - mass_remain, 1)
        self.noaa_mass_balance['mass_components'][surfaceID, :] = \
            mass_remain
        self.elements.mass_oil[surface] = np.sum(mass_remain, 1)

    def emulsification_noaa(self):
        #############################################
        # Emulsification (surface only?)
        #############################################
        logger.debug('    Calculating emulsification - NOAA')
        emul_time = self.oiltype.bulltime
        emul_constant = self.oiltype.bullwinkle
        # max water content fraction - get from database
        Y_max = self.oiltype.emulsion_water_fraction_max
        if self.oil_name in self.max_water_fraction:
            max_water_fraction = self.max_water_fraction[self.oil_name]
            logger.debug(
                'Overriding max water fraxtion with value %f instead of default %f'
                % (max_water_fraction, Y_max))
            Y_max = max_water_fraction
        # emulsion
        if Y_max <= 0:
            logger.debug('Oil does not emulsify, returning.')
            return
        # Constants for droplets
        drop_min = 1.0e-6
        drop_max = 1.0e-5
        S_max = (6. / drop_min) * (Y_max / (1.0 - Y_max))
        S_min = (6. / drop_max) * (Y_max / (1.0 - Y_max))
        # Emulsify...
        fraction_evaporated = self.elements.mass_evaporated / (
            self.elements.mass_evaporated + self.elements.mass_oil)
        # f ((le_age >= emul_time && emul_time >= 0.) || frac_evap[i] >= emul_C && emul_C > 0.)

        start_emulsion = np.where((
            (self.elements.age_seconds >= emul_time) & (emul_time >= 0))
                                  | ((fraction_evaporated >= emul_constant)
                                     & (emul_constant > 0)))[0]
        if len(start_emulsion) == 0:
            logger.debug('        Emulsification not yet started')
            return

        if self.oiltype.bulltime > 0:  # User has set value
            start_time = self.oiltype.bulltime * np.ones(len(start_emulsion))
        else:
            start_time = self.elements.age_seconds[start_emulsion]
            # TODO: it should be possible to specify oil age at seeding
            start_time[self.elements.age_seconds[start_emulsion] >=
                       0] = self.elements.bulltime[start_emulsion]
        # Update droplet interfacial area
        k_emul = noaa.water_uptake_coefficient(
            self.oiltype,
            self.wind_speed()[start_emulsion])
        self.elements.interfacial_area[start_emulsion] = \
            self.elements.interfacial_area[start_emulsion] + \
            (k_emul*self.time_step.total_seconds()*
             np.exp((-k_emul/S_max)*(
                self.elements.age_seconds[start_emulsion] - start_time)))
        self.elements.interfacial_area[
            self.elements.interfacial_area > S_max] = S_max
        # Update water fraction
        self.elements.water_fraction[start_emulsion] = (
            self.elements.interfacial_area[start_emulsion] * drop_max /
            (6.0 +
             (self.elements.interfacial_area[start_emulsion] * drop_max)))
        self.elements.water_fraction[self.elements.interfacial_area >= (
            (6.0 / drop_max) * (Y_max / (1.0 - Y_max)))] = Y_max

    def update_terminal_velocity(self,
                                 Tprofiles=None,
                                 Sprofiles=None,
                                 z_index=None):
        """Calculate terminal velocity for oil droplets

        according to:

        * Tkalich et al. (2002): Vertical mixing of oil droplets by breaking waves

        * Marine Pollution Bulletin 44, 1219-1229

        If profiles of temperature and salt are passed into this function,
        they will be interpolated from the profiles.
        if not, T,S will be fetched from reader.
        """
        g = 9.81  # ms-2

        r = self.elements.diameter  # NB: r is diameter, not radius

        # Prepare interpolation of temp, salt

        if not (Tprofiles is None and Sprofiles is None):
            if z_index is None:
                z_i = range(Tprofiles.shape[0])  # evtl. move out of loop
                # evtl. move out of loop
                z_index = interp1d(-self.environment_profiles['z'],
                                   z_i,
                                   bounds_error=False)
            zi = z_index(-self.elements.z)
            upper = np.maximum(np.floor(zi).astype(np.uint8), 0)
            lower = np.minimum(upper + 1, Tprofiles.shape[0] - 1)
            weight_upper = 1 - (zi - upper)

        # Do interpolation of temp, salt if profiles were passed into
        # this function, if not, use reader by calling self.environment
        if Tprofiles is None:
            T0 = self.environment.sea_water_temperature
        else:
            T0 = Tprofiles[upper, range(Tprofiles.shape[1])] * \
                weight_upper + \
                Tprofiles[lower, range(Tprofiles.shape[1])] * \
                (1-weight_upper)
        if Sprofiles is None:
            S0 = self.environment.sea_water_salinity
        else:
            S0 = Sprofiles[upper, range(Sprofiles.shape[1])] * \
                weight_upper + \
                Sprofiles[lower, range(Sprofiles.shape[1])] * \
                (1-weight_upper)

        rho_oil = self.elements.density
        rho_water = self.sea_water_density(T=T0, S=S0)

        # dynamic water viscosity
        my_w = 0.001 * (1.7915 - 0.0538 * T0 + 0.007 *
                        (T0**(2.0)) - 0.0023 * S0)
        # ~0.0014 kg m-1 s-1
        # kinemativ water viscosity
        ny_w = my_w / rho_water
        rhopr = rho_oil / rho_water

        # terminal velocity for low Reynolds numbers
        kw = 2 * g * (1 - rhopr) / (9 * ny_w)
        W = kw * r**2

        # check if we are in a high Reynolds number regime
        Re = 2 * r * W / ny_w
        highRe = np.where(Re > 50)

        # Terminal velocity in high Reynolds numbers
        kw = (16 * g * (1 - rhopr) / 3)**0.5
        W2 = kw * r**0.5

        W[highRe] = W2[highRe]
        self.elements.terminal_velocity = W

    def oil_wave_entrainment_rate(self):
        er = self.get_config('wave_entrainment:entrainment_rate')
        if er == 'Li et al. (2017)':
            entrainment_rate = oil_wave_entrainment_rate_li2017(
                dynamic_viscosity=self.elements.viscosity *
                self.elements.density,
                oil_density=self.elements.density,
                interfacial_tension=self.oil_water_interfacial_tension,
                significant_wave_height=self.significant_wave_height(),
                wave_breaking_fraction=self.sea_surface_wave_breaking_fraction(
                ),
                sea_water_density=self.sea_water_density())
            return entrainment_rate
        else:
            logger.error("no entrainment rate mechanism configured")
            raise Exception("no entrainment rate mechanism configured")

    def prepare_vertical_mixing(self):
        '''Calculate entrainment probability before main loop'''
        self.oil_entrainment_probability = \
            1 - np.exp(-self.oil_wave_entrainment_rate()*\
                       self.get_config('vertical_mixing:timestep'))
        # Calculate a random droplet diameter for each particle,
        # to be used if this particle gets entrained
        self.droplet_diameter_if_entrained = \
            self.get_wave_breaking_droplet_diameter()
        # Uncomment lines below to plot droplet size distribution at each step
        #import matplotlib.pyplot as plt
        #plt.hist(self.droplet_diameter_if_entrained, 200)
        #plt.gca().set_xscale("log")
        #plt.gca().set_yscale("log")
        #plt.show()

    def surface_wave_mixing(self, time_step_seconds):
        """Mix surface oil into water column."""
        # Entrain oil into uppermost layer (whitecapping from waves)
        # TODO: optimise this by only calculate for surface elements
        surface = self.elements.z >= 0
        random_number = np.random.uniform(0, 1, len(self.elements.z))
        entrained = np.logical_and(
            surface, random_number < self.oil_entrainment_probability)

        # Intrusion depth for wave entrainment from
        # Delvigne and Sweeney (1988), Li et al. (2017):
        if entrained.sum() > 0:
            logger.debug('Entraining %i of %i surface elements' %
                         (entrained.sum(), surface.sum()))
            zb = 1.5 * self.significant_wave_height()  # between 0 and zb
            intrusion_depth = np.random.uniform(0, np.mean(zb),
                                                entrained.sum())
            self.elements.z[entrained] = -intrusion_depth
            if self.keep_droplet_diameter is False:
                # Give entrained elements a random diameter
                self.elements.diameter[entrained] = \
                    self.droplet_diameter_if_entrained[entrained]

    def surface_stick(self):
        """set surfaced particles to exactly zero depth to let them form a slick """

        surface = np.where(self.elements.z >= 0)
        if len(surface[0]) > 0:
            self.elements.z[surface] = 0.

    def get_wave_breaking_droplet_diameter(self):
        dm = self.get_config('wave_entrainment:droplet_size_distribution')
        if dm == 'Johansen et al. (2015)':
            return self.get_wave_breaking_droplet_diameter_johansen2015()
        elif dm == 'Li et al. (2017)':
            return self.get_wave_breaking_droplet_diameter_liz2017()
        else:
            raise Exception("no droplet size distribution specified")

    def get_wave_breaking_droplet_diameter_liz2017(self):
        # Li,Zhengkai, M. Spaulding, D. French-McCay, D. Crowley, J.R. Payne: "Development of a unified oil droplet size distribution model
        # with application to surface breaking waves and subsea blowout releases considering dispersant effects" Mar. Pol. Bul.
        # DOI: 10.1016/j.marpolbul.2016.09.008
        # Should be prefered when the oil film thickness is unknown.
        if not hasattr(self, 'droplet_spectrum_pdf'):
            # Generate droplet spectrum as in Li (Zhengkai) et al. (2017)
            # Bounds are hardcoded to 1 micron and 3mm
            logger.debug('Generating wave breaking droplet size spectrum')
            self.droplet_spectrum_diameter = np.linspace(1e-6, 3e-3, 1000000)
            g = 9.81
            interfacial_tension = self.oil_water_interfacial_tension
            delta_rho = self.sea_water_density() - self.elements.density
            d_o = 4 * (interfacial_tension / (delta_rho * g))**0.5
            we = (self.sea_water_density() * g *
                  self.significant_wave_height() * d_o) / interfacial_tension
            oh = self.elements.viscosity * self.elements.density * (
                self.elements.density * interfacial_tension *
                d_o)**-0.5  # From kin. to dyn. viscosity by * density
            r = 1.791
            p = 0.460
            q = -0.518
            dV_50 = d_o * r * (
                1 + 10 * oh
            )**p * we**q  # median droplet diameter in volume distribution
            sd = 0.4  # log standard deviation in log10 units
            Sd = np.log(10) * sd  # log standard deviation in natural log units
            # TODO: calculation below with scalars, but we have arrays, with varying oil properties
            # treat all particle in one go:
            dV_50 = np.mean(dV_50)  # mean log diameter
            dN_50 = np.exp(
                np.log(dV_50) - 3 *
                Sd**2)  # convert number distribution to volume distribution
            logger.debug(
                'Droplet distribution median diameter dV_50: %f, dN_50: %f ' %
                (dV_50, np.mean(dN_50)))
            spectrum = (np.exp(
                -(np.log(self.droplet_spectrum_diameter) - np.log(dV_50))**2 /
                (2 * Sd**2))) / (self.droplet_spectrum_diameter * Sd *
                                 np.sqrt(2 * np.pi))
            self.droplet_spectrum_pdf = spectrum / np.sum(spectrum)
        if ~np.isfinite(np.sum(self.droplet_spectrum_pdf)) or \
                np.abs(np.sum(self.droplet_spectrum_pdf) - 1) > 1e-6:
            logger.warning('Could not update droplet diameters.')
            return self.elements.diameter
        else:
            return np.random.choice(self.droplet_spectrum_diameter,
                                    size=self.num_elements_active(),
                                    p=self.droplet_spectrum_pdf)

    def get_wave_breaking_droplet_diameter_johansen2015(self):
        # Johansen O, Reed M, Bodsberg NR, Natural dispersion revisited
        # DOI: 10.1016/j.marpolbul.2015.02.026
        # requires oil film thickness
        if not hasattr(self, 'droplet_spectrum_pdf') or self.get_config(
                'processes:update_oilfilm_thickness') is True:
            # Generate droplet spectrum as in Johansen et al. (2015)
            # Bounds are hardcoded to 1micron and 3mm
            logger.debug('Generating wave breaking droplet size spectrum')
            self.droplet_spectrum_diameter = np.linspace(1e-6, 3e-3, 1000000)
            g = 9.81
            interfacial_tension = self.oil_water_interfacial_tension
            H = self.significant_wave_height(
            )  # fall height = 2 * wave amplitude
            # Reyolds number (Eq. 7a from Johansen et al. 2015)
            re = (self.elements.density * self.elements.oil_film_thickness *
                  (g * H)**0.5) / (self.elements.viscosity *
                                   self.elements.density)
            # Weber number (Eq. 7b from Johansen et al.2015)
            we = (self.elements.density * self.elements.oil_film_thickness *
                  g * H) / interfacial_tension  # Weber number
            A = 2.251  # parameters from Johansen et al. 2015
            Bp = 0.027
            B = A * Bp
            dN_50 = (A * self.elements.oil_film_thickness * we**-0.6) + (
                B * self.elements.oil_film_thickness * re**-0.6)
            # median droplet diameter in number distribution
            sd = 0.4  # log standard deviation in log10 units
            Sd = np.log(10) * sd  # log standard deviation in natural log units
            # Convert number distribution to volume distribution
            dV_50 = np.exp(np.log(dN_50) + 3 * Sd**2)
            # TODO: calculation below with scalars, but we have
            # arrays, with varying oil properties
            # treat all particle in one go:
            dV_50 = np.mean(dV_50)  # mean log diameter
            logger.debug(
                'Droplet distribution median diameter dV_50: %f, dN_50: %f ' %
                (dV_50, np.mean(dN_50)))
            spectrum = (np.exp(
                -(np.log(self.droplet_spectrum_diameter) - np.log(dV_50))**2 /
                (2 * Sd**2))) / (self.droplet_spectrum_diameter * Sd *
                                 np.sqrt(2 * np.pi))
            self.droplet_spectrum_pdf = spectrum / np.sum(spectrum)
        if ~np.isfinite(np.sum(self.droplet_spectrum_pdf)) or \
                np.abs(np.sum(self.droplet_spectrum_pdf) - 1) > 1e-6:
            logger.warning('Could not update droplet diameters.')
            return self.elements.diameter
        else:
            return np.random.choice(self.droplet_spectrum_diameter,
                                    size=self.num_elements_active(),
                                    p=self.droplet_spectrum_pdf)

    def resurface_elements(self, minimum_depth=None):
        """Oil elements reaching surface (or above) form slick, not droplet"""
        surface = np.where(self.elements.z >= 0)[0]
        self.elements.z[surface] = 0

    def advect_oil(self):

        # Calculating various drift factors according to ice concentration
        if hasattr(self.environment, 'sea_ice_area_fraction'):
            A = self.environment.sea_ice_area_fraction
            # According to
            # Nordam T, Beegle-Krause CJ, Skancke J, Nepstad R, Reed M.
            # Improving oil spill trajectory modelling in the Arctic.
            # Mar Pollut Bull. 2019;140:65-74.
            # doi:10.1016/j.marpolbul.2019.01.019
            k_ice = (A - 0.3) / (0.8 - 0.3)
            k_ice[A < 0.3] = 0
            k_ice[A > 0.8] = 1
            if k_ice.max() > 0:
                logger.info(
                    'Ice concentration above 30%, using Nordam scheme for advection in ice'
                )
            # Using decreased Stokes drift according to
            # Arneborg, L. (2017). Oil drift modellling in pack ice
            # - Sensitivity of oil-in-ice parameters.
            # Ocean Engineering 144 (2017) 340-350
            factor_stokes = (0.7 - A) / 0.7
            factor_stokes[A > 0.7] = 0
        else:
            k_ice = 0
            factor_stokes = 1

        # Simply move particles with ambient current
        self.advect_ocean_current(factor=1 - k_ice)

        # Wind drag for elements at ocean surface
        self.advect_wind(factor=1 - k_ice)

        # Stokes drift
        self.stokes_drift(factor_stokes)

        # Advect with ice
        self.advect_with_sea_ice(factor=k_ice)

    def update(self):
        """Update positions and properties of oil particles."""

        # TODO: move all config-checking inside respective methods
        if self.get_config('processes:update_oilfilm_thickness') is True:
            self.update_surface_oilfilm_thickness()

        # Oil weathering (inherited from OpenOil)
        self.oil_weathering()

        # Turbulent Mixing
        if self.get_config('drift:vertical_mixing') is True:
            self.update_terminal_velocity()
            self.vertical_mixing()
            del self.droplet_spectrum_pdf

        # Vertical advection
        if self.get_config('drift:vertical_advection') is True:
            self.vertical_advection()

        # Horizontal advection (inherited from OpenOil)
        self.advect_oil()

    def get_oil_budget(self):
        """Get oil budget for the current simulation

        The oil budget consists of the following categories:

        * surface: the sum of variable mass_oil for all active elements where z = 0
        * submerged: the sum of variable mass_oil for all active elements where z < 0
        * stranded: the sum of variable mass_oil for all elements which are stranded
        * evaporated: the sum of variable mass_evaporated for all elements
        * dispersed: the sum of variable mass_dispersed for all elements

        The sum (total mass) shall equal the mass released. Note that the mass of oil
        is conserved, whereas the volume may change with changes in density and
        water uptake (emulsification). Therefore mass should be used for budgets,
        eventually converted to volume (by dividing on density) in the final step
        before presentation.

        Note that mass_oil is the mass of pure oil. The mass of oil emulsion
        (oil containing entrained water droplets) can be calculated as:

        .. code::

            mass_emulsion = mass_oil / (1 - water_fraction)

        I.e. water_fraction = 0 means pure oil, water_fraction = 0.5 means mixture of
        50% oil and 50% water, and water_fraction = 0.9 (which is maximum)
        means 10% oil and 90% water.
        """

        if self.time_step.days < 0:  # Backwards simulation
            return None

        z, dummy = self.get_property('z')
        mass_oil, status = self.get_property('mass_oil')
        density = self.get_property('density')[0][0, 0]

        if 'stranded' not in self.status_categories:
            self.status_categories.append('stranded')
        mass_submerged = np.ma.masked_where(
            ((status == self.status_categories.index('stranded')) |
             (z == 0.0)), mass_oil)
        mass_submerged = np.ma.sum(mass_submerged, axis=1).filled(0)

        mass_surface = np.ma.masked_where(
            ((status == self.status_categories.index('stranded')) | (z < 0.0)),
            mass_oil)
        mass_surface = np.ma.sum(mass_surface, axis=1).filled(0)

        mass_stranded = np.ma.sum(np.ma.masked_where(
            status != self.status_categories.index('stranded'), mass_oil),
                                  axis=1).filled(0)
        mass_evaporated, status = self.get_property('mass_evaporated')
        mass_evaporated = np.sum(mass_evaporated, axis=1).filled(0)
        mass_dispersed, status = self.get_property('mass_dispersed')
        mass_dispersed = np.sum(mass_dispersed, axis=1).filled(0)
        mass_biodegraded, status = self.get_property('mass_biodegraded')
        mass_biodegraded = np.sum(mass_biodegraded, axis=1).filled(0)

        oil_budget = {
            'oil_density':
            density,
            'mass_dispersed':
            mass_dispersed,
            'mass_submerged':
            mass_submerged,
            'mass_surface':
            mass_surface,
            'mass_stranded':
            mass_stranded,
            'mass_evaporated':
            mass_evaporated,
            'mass_biodegraded':
            mass_biodegraded,
            'mass_total': (mass_dispersed + mass_submerged + mass_surface +
                           mass_stranded + mass_evaporated + mass_biodegraded)
        }

        return oil_budget

    def plot_oil_budget(self,
                        filename=None,
                        ax=None,
                        show_density_viscosity=True,
                        show_wind_and_current=True):

        if self.time_step.days < 0:  # Backwards simulation
            fig = plt.figure(figsize=(10, 6.))
            plt.text(0.1, 0.5, 'Oil weathering deactivated for '
                     'backwards simulations')
            plt.axis('off')
            if filename is not None:
                plt.savefig(filename)
                plt.close()
            else:
                plt.show()
            return

        b = self.get_oil_budget()

        oil_budget = np.row_stack(
            (b['mass_dispersed'], b['mass_submerged'], b['mass_surface'],
             b['mass_stranded'], b['mass_evaporated'], b['mass_biodegraded']))
        oil_density = b['oil_density']

        budget = np.cumsum(oil_budget, axis=0)

        time, time_relative = self.get_time_array()
        time = np.array([t.total_seconds() / 3600. for t in time_relative])

        if ax is None:
            # Left axis showing oil mass
            nrows = 1
            if show_density_viscosity is True:
                nrows = nrows + 1
            if show_wind_and_current is True:
                nrows = nrows + 1
            fig, axs = plt.subplots(
                nrows=nrows, ncols=1,
                figsize=(10, 6. + (nrows - 1) * 3))  # Suitable aspect ratio
            #ax1 = fig.add_subplot(nrows=nrows, 1, 1)
            if nrows == 1:
                ax1 = axs
            elif nrows >= 2:
                ax1 = axs[0]
                if show_density_viscosity is True:
                    self.plot_oil_density_and_viscosity(ax=axs[1], show=False)
                if show_wind_and_current is True:
                    self.plot_environment(ax=axs[nrows - 1], show=False)
        else:
            ax1 = ax

        # Hack: make some emply plots since fill_between does not support label
        if np.sum(b['mass_dispersed']) > 0:
            ax1.add_patch(
                plt.Rectangle((0, 0),
                              0,
                              0,
                              color='darkslategrey',
                              label='dispersed'))
            ax1.fill_between(time, 0, budget[0, :], facecolor='darkslategrey')
        if np.sum(b['mass_submerged']) > 0:
            ax1.add_patch(
                plt.Rectangle((0, 0),
                              0,
                              0,
                              color='darkblue',
                              label='submerged'))
            ax1.fill_between(time,
                             budget[0, :],
                             budget[1, :],
                             facecolor='darkblue')
        if np.sum(b['mass_surface']) > 0:
            ax1.add_patch(
                plt.Rectangle((0, 0), 0, 0, color='royalblue',
                              label='surface'))
            ax1.fill_between(time,
                             budget[1, :],
                             budget[2, :],
                             facecolor='royalblue')
        if np.sum(b['mass_stranded']) > 0:
            ax1.add_patch(
                plt.Rectangle((0, 0), 0, 0, color='black', label='stranded'))
            ax1.fill_between(time,
                             budget[2, :],
                             budget[3, :],
                             facecolor='black')
        if np.sum(b['mass_evaporated']) > 0:
            ax1.add_patch(
                plt.Rectangle((0, 0),
                              0,
                              0,
                              color='skyblue',
                              label='evaporated'))
            ax1.fill_between(time,
                             budget[3, :],
                             budget[4, :],
                             facecolor='skyblue')
        if np.sum(b['mass_biodegraded']) > 0:
            ax1.add_patch(
                plt.Rectangle((0, 0),
                              0,
                              0,
                              color='indigo',
                              label='biodegraded'))
            ax1.fill_between(time,
                             budget[4, :],
                             budget[5, :],
                             facecolor='indigo')

        ax1.set_ylim([0, budget.max()])
        ax1.set_xlim([0, time.max()])
        ax1.set_ylabel('Mass oil  [%s]' %
                       self.elements.variables['mass_oil']['units'])
        ax1.set_xlabel('Time  [hours]')
        # Right axis showing volume
        ax2 = ax1.twinx()
        mass_total = b['mass_total'][-1]
        ax2.set_ylim([0, mass_total / oil_density])
        ax2.set_ylabel('Volume oil [m3]')
        plt.title('%s (%.1f kg/m3) - %s to %s' %
                  (self.get_oil_name(), oil_density,
                   self.start_time.strftime('%Y-%m-%d %H:%M'),
                   self.time.strftime('%Y-%m-%d %H:%M')))
        # Shrink current axis's height by 10% on the bottom
        box = ax1.get_position()
        ax1.set_position(
            [box.x0, box.y0 + box.height * 0.1, box.width, box.height * 0.9])
        ax2.set_position(
            [box.x0, box.y0 + box.height * 0.1, box.width, box.height * 0.9])
        ax1.legend(bbox_to_anchor=(0., -0.10, 1., -0.03),
                   loc=1,
                   ncol=6,
                   mode="expand",
                   borderaxespad=0.,
                   fontsize=10)
        if filename is not None:
            plt.savefig(filename)
            plt.close()
        else:
            plt.show()

    def get_oil_name(self):
        try:
            return self.get_config('seed:oil_type')
        except:  # fallback if importing old files
            return 'unknown oiltype'

    def cumulative_oil_entrainment_fraction(self):
        '''Returns the fraction of oil elements which has been entrained vs time'''
        z = self.get_property('z')[0].copy()
        z = np.ma.masked_where(z == 0, z)
        me = np.ma.notmasked_edges(z, axis=0)
        maskfirst = me[0][0]
        maskrow = me[0][1]
        z = z * 0
        for mf, mr in zip(maskfirst, maskrow):
            z[mf:z.shape[0], mr] = 1  # has been entrained
        totentrained = np.sum(z, 1)
        cumulative_fraction_entrained = np.sum(z, 1) / z.shape[1]
        return cumulative_fraction_entrained

    def plot_oil_density_and_viscosity(self, ax=None, show=True):
        if ax is None:
            fig, ax = plt.subplots()
        import matplotlib.dates as mdates

        time, time_relative = self.get_time_array()
        time = np.array([t.total_seconds() / 3600. for t in time_relative])
        kin_viscosity = self.history['viscosity']
        dyn_viscosity = kin_viscosity * self.history['density']
        dyn_viscosity_mean = dyn_viscosity.mean(axis=0)
        dyn_viscosity_std = dyn_viscosity.std(axis=0)
        density = self.history['density'].mean(axis=0)
        density_std = self.history['density'].std(axis=0)

        ax.plot(time,
                dyn_viscosity_mean,
                'g',
                lw=2,
                label='Dynamical viscosity')
        ax.fill_between(time,
                        dyn_viscosity_mean - dyn_viscosity_std,
                        dyn_viscosity_mean + dyn_viscosity_std,
                        color='g',
                        alpha=0.5)
        ax.set_ylim([0, max(dyn_viscosity_mean + dyn_viscosity_std)])
        ax.set_ylabel(r'Dynamical viscosity  [cPoise] / [mPas]', color='g')
        ax.tick_params(axis='y', colors='g')

        axb = ax.twinx()
        axb.plot(time, density, 'b', lw=2, label='Density')
        axb.fill_between(time,
                         density - density_std,
                         density + density_std,
                         color='b',
                         alpha=0.5)
        ax.set_xlim([0, time.max()])
        ax.set_xlabel('Time [hours]')
        axb.set_ylabel(r'Density  [kg/m3]', color='b')
        axb.tick_params(axis='y', colors='b')

        ax.legend(loc='upper left')
        axb.legend(loc='lower right')
        if show is True:
            plt.show()

    def set_oiltype(self, oiltype):
        oiltype = adios.oil_name_alias.get(oiltype, oiltype)
        logger.info(f'setting oil_type to: {oiltype}')

        self.oil_name = oiltype
        self.set_config('seed:oil_type', oiltype)
        if self.oil_weathering_model == 'noaa':
            self.oiltype = adios.find_full_oil_from_name(self.oil_name)
            if not self.oiltype.valid():
                logger.error(
                    f"{self.oiltype} is not a valid oil for Opendrift simulations"
                )
                raise ValueError()
        else:
            raise ValueError("unsupported oil weathering model")

    def set_oiltype_by_id(self, oiltypeid):
        if self.oil_weathering_model == 'noaa':
            self.oiltype = adios.get_full_oil_from_id(oiltypeid)
            self.oil_name = self.oiltype.name
            if not self.oiltype.valid():
                logger.error(
                    f"{self.oiltype} is not a valid oil for Opendrift simulations"
                )
                raise ValueError()
        else:
            raise ValueError("unsupported oil weathering model")

    def set_oiltype_by_json(self, json):
        if self.oil_weathering_model == 'noaa':
            self.oiltype = adios.oil.OpendriftOil(json)
            self.oil_name = self.oiltype.name
            if not self.oiltype.valid():
                logger.error(
                    f"{self.oiltype} is not a valid oil for Opendrift simulations"
                )
                raise ValueError()
        else:
            raise ValueError("unsupported oil weathering model")

    def set_oiltype_from_file(self, path):
        if self.oil_weathering_model == 'noaa':
            import json
            with open(path, 'r') as fd:
                j = json.load(fd)

            self.oiltype = adios.oil.OpendriftOil(j)
            self.oil_name = self.oiltype.name
            if not self.oiltype.valid():
                logger.error(
                    f"{self.oiltype} is not a valid oil for Opendrift simulations"
                )
                raise ValueError()
        else:
            raise ValueError("unsupported oil weathering model")

    def store_oil_seed_metadata(self, **kwargs):
        for s in [
                'lon', 'lat', 'radius', 'time', 'number', 'z', 'm3_per_hour'
        ]:
            if not 'seed_' + s in self.metadata_dict:
                if s in kwargs:
                    val = kwargs[s]
                else:
                    if s == 'radius':
                        val = 0  # There is no default radius
                    elif s == 'z' and 'z' not in kwargs and self.get_config(
                            'seed:seafloor') is True:
                        val = 'seafloor'
                    else:
                        val = self.get_config('seed:' + s)
                if s == 'time':
                    if hasattr(kwargs[s], '__len__'):
                        self.add_metadata('seed_time', val[0])
                    else:
                        self.add_metadata('seed_time', val)
                elif isinstance(val, str):
                    self.add_metadata('seed_' + s, val)
                else:
                    self.add_metadata('seed_' + s, np.atleast_1d(val).mean())
        if not 'seed_oiltype' in self.metadata_dict:
            if 'oiltype' in kwargs:
                oiltype = kwargs['oiltype']
            elif 'oil_type' in kwargs:
                oiltype = kwargs['oil_type']
            else:
                oiltype = self.get_config('seed:oil_type')
            self.add_metadata('seed_oiltype', oiltype)

    def seed_elements(self, *args, **kwargs):

        if len(args) == 2:
            kwargs['lon'] = args[0]
            kwargs['lat'] = args[1]
            args = {}

        self.store_oil_seed_metadata(**kwargs)

        if 'number' not in kwargs:
            number = self.get_config('seed:number')
        else:
            number = kwargs['number']
        if 'diameter' in kwargs:
            logger.info('Droplet diameter is provided, and will '
                        'be kept constant during simulation')
            self.keep_droplet_diameter = True
        else:
            self.keep_droplet_diameter = False
        if 'z' not in kwargs or kwargs['z'] is None:
            if self.get_config('seed:seafloor') is True:
                kwargs['z'] = 'seafloor'
            else:
                kwargs['z'] = self.get_config('seed:z')
        if isinstance(kwargs['z'], str) and \
                kwargs['z'][0:8] == 'seafloor':
            z = -np.ones(number)
        else:
            z = np.atleast_1d(kwargs['z'])
        if len(z) == 1:
            z = z * np.ones(number)  # Convert scalar z to array
        subsea = z < 0
        if np.sum(subsea) > 0 and 'diameter' not in kwargs:
            # Droplet min and max for particles seeded below sea surface
            sub_dmin = self.get_config('seed:droplet_diameter_min_subsea')
            sub_dmax = self.get_config('seed:droplet_diameter_max_subsea')
            logger.info('Using particle diameters between %s and %s m for '
                        'elements seeded below sea surface.' %
                        (sub_dmin, sub_dmax))
            kwargs['diameter'] = np.random.uniform(sub_dmin, sub_dmax, number)

        if 'oiltype' in kwargs:
            logger.warning(
                'Seed argument *oiltype* is deprecated, use *oil_type* instead'
            )
            kwargs['oil_type'] = kwargs['oiltype']
            del kwargs['oiltype']

        if 'oil_type' in kwargs:
            self.set_config('seed:oil_type', kwargs['oil_type'])
            del kwargs['oil_type']
        else:
            logger.info('Oil type not specified, using default: ' +
                        self.get_config('seed:oil_type'))
        self.set_oiltype(self.get_config('seed:oil_type'))

        if self.oil_weathering_model == 'noaa':
            oil_density = self.oiltype.density_at_temp(285)
            oil_viscosity = self.oiltype.kvis_at_temp(285)
            logger.info(
                'Using density %s and viscosity %s of oiltype %s' %
                (oil_density, oil_viscosity, self.get_config('seed:oil_type')))
            kwargs['density'] = oil_density
            kwargs['viscosity'] = oil_viscosity

        if 'm3_per_hour' in kwargs:
            m3_per_hour = kwargs['m3_per_hour']
            del kwargs['m3_per_hour']
        else:
            m3_per_hour = self.get_config('seed:m3_per_hour')

        if 'number' in kwargs:
            num_elements = kwargs['number']
        else:
            num_elements = self.get_config('seed:number')
        time = kwargs['time']
        if type(time) is list:
            duration_hours = ((time[1] - time[0]).total_seconds()) / 3600
            if duration_hours == 0:
                duration_hours = 1.
        else:
            duration_hours = 1.  # For instantaneous spill, we use 1h
        kwargs['mass_oil'] = (m3_per_hour * duration_hours / num_elements *
                              kwargs['density'])

        super(OpenOil, self).seed_elements(*args, **kwargs)

    def seed_cone(self, *args, **kwargs):
        if 'oiltype' in kwargs:
            logger.warning(
                'Seed argument *oiltype* is deprecated, use *oil_type* instead'
            )
            kwargs['oil_type'] = kwargs['oiltype']
            del kwargs['oiltype']

        self.store_oil_seed_metadata(**kwargs)

        super(OpenOil, self).seed_cone(*args, **kwargs)

    def seed_from_gml(self, gmlfile, num_elements=1000, *args, **kwargs):
        """Read oil slick contours from GML file, and seed particles within."""

        # Specific imports
        import datetime
        from matplotlib.path import Path
        from xml.etree import ElementTree
        from matplotlib.patches import Polygon
        import pyproj

        namespaces = {
            'od': 'http://cweb.ksat.no/cweb/schema/geoweb/oil',
            'gml': 'http://www.opengis.net/gml'
        }
        slicks = []

        with open(gmlfile, 'rt') as e:
            tree = ElementTree.parse(e)

        pos1 = 'od:oilDetectionMember/od:oilDetection/od:oilSpill/gml:Polygon'
        pos2 = 'gml:exterior/gml:LinearRing/gml:posList'

        # This retrieves some other types of patches, found in some files only
        # Should be combines with the above, to get all patches
        #pos1 = 'od:oilDetectionMember/od:oilDetection/od:oilSpill/
        # gml:Surface/gml:polygonPatches'
        #pos2 = 'gml:PolygonPatch/gml:exterior/gml:LinearRing/gml:posList'

        # Find detection time
        time_pos = 'od:oilDetectionMember/od:oilDetection/od:detectionTime'
        oil_time = datetime.datetime.strptime(
            tree.find(time_pos, namespaces).text, '%Y-%m-%dT%H:%M:%S.%fZ')

        for patch in tree.findall(pos1, namespaces):
            pos = patch.find(pos2, namespaces).text
            c = np.array(pos.split()).astype(float)
            lon = c[0::2]
            lat = c[1::2]
            slicks.append(Polygon(list(zip(lon, lat))))

        # Find boundary and area of all patches
        lons = np.array([])
        lats = lons.copy()
        for slick in slicks:
            ext = slick.get_extents()
            lons = np.append(lons, [ext.xmin, ext.xmax])
            lats = np.append(lats, [ext.ymin, ext.ymax])
            # Make a stereographic projection centred on the polygon
        lonmin = lons.min()
        lonmax = lons.max()
        latmin = lats.min()
        latmax = lats.max()

        # Place n points within the polygons
        proj = pyproj.Proj(
            '+proj=aea +lat_1=%f +lat_2=%f +lat_0=%f +lon_0=%f' %
            (latmin, latmax, (latmin + latmax) / 2, (lonmin + lonmax) / 2))
        slickarea = np.array([])
        for slick in slicks:
            lonlat = slick.get_xy()
            lon = lonlat[:, 0]
            lat = lonlat[:, 1]
            x, y = proj(lon, lat)

            area_of_polygon = 0.0
            for i in range(-1, len(x) - 1):
                area_of_polygon += x[i] * (y[i + 1] - y[i - 1])
            area_of_polygon = abs(area_of_polygon) / 2.0
            slickarea = np.append(slickarea, area_of_polygon)  # in m2

        # Make points
        deltax = np.sqrt(np.sum(slickarea) / num_elements)

        lonpoints = np.array([])
        latpoints = np.array([])
        for i, slick in enumerate(slicks):
            lonlat = slick.get_xy()
            lon = lonlat[:, 0]
            lat = lonlat[:, 1]
            x, y = proj(lon, lat)
            xvec = np.arange(x.min(), x.max(), deltax)
            yvec = np.arange(y.min(), y.max(), deltax)
            x, y = np.meshgrid(xvec, yvec)
            lon, lat = proj(x, y, inverse=True)
            lon = lon.ravel()
            lat = lat.ravel()
            points = np.c_[lon, lat]
            ind = Path(slick.xy).contains_points(points)
            lonpoints = np.append(lonpoints, lon[ind])
            latpoints = np.append(latpoints, lat[ind])

        # Finally seed at found positions
        kwargs['lon'] = lonpoints
        kwargs['lat'] = latpoints
        self.seed_elements(time=oil_time, **kwargs)

    def seed_from_geotiff_thickness(self,
                                    filename,
                                    number=50000,
                                    *args,
                                    **kwargs):
        '''Seed from files as provided by Prof. Chuanmin Hu'''

        from osgeo import gdal, ogr, osr

        if not 'time' in kwargs:
            try:  # get time from filename
                time = datetime.strptime(filename[-28:-13], '%Y%m%d.%H%M%S')
                logger.info('Parsed time from filename: %s' % time)
            except:
                time = datetime.now()
                logger.warning('Could not pase time from filename, '
                               'using present time: %s' % time)

        ds = gdal.Open(filename)

        srcband = ds.GetRasterBand(1)
        data = srcband.ReadAsArray()

        thickness_microns = [0.04, 0.44, 4.4, 16]  # NB: approximate
        categories = [1, 2, 3, 4]  # categories

        # Make memory raster bands for each category
        memrastername = filename + '.mem'
        memdriver = gdal.GetDriverByName('MEM')
        mem_ds = memdriver.CreateCopy(memrastername, ds)
        for cat in categories:
            mem_ds.AddBand(gdal.GDT_Byte)
            mem_band = mem_ds.GetRasterBand(cat)
            mem_band.WriteArray(data == cat)

        # Make memory polygons for each category
        drv = ogr.GetDriverByName('MEMORY')
        mem_vector_ds = [0] * len(categories)
        mem_vector_layers = [0] * len(categories)
        for cat in categories:
            memshapename = filename + '%i.mem' % cat
            mem_vector_ds[cat - 1] = drv.CreateDataSource(memshapename)
            mem_vector_layers[cat-1] = \
                mem_vector_ds[cat-1].CreateLayer(
                    'thickness%i' % cat, srs=None)
            gdal.Polygonize(mem_ds.GetRasterBand(cat),
                            None,
                            mem_vector_layers[cat - 1],
                            -1, [],
                            callback=None)

        total_area = np.zeros(len(categories))
        layers = [0] * len(categories)

        src_srs = osr.SpatialReference()
        src_srs.ImportFromEPSG(4269)
        tgt_srs = osr.SpatialReference()
        tgt_srs.ImportFromEPSG(3857)
        transform = osr.CoordinateTransformation(src_srs, tgt_srs)
        for cat in categories:
            memshapename = filename + '%i.shp' % cat
            layers[cat - 1] = mem_vector_layers[cat - 1]
            areas = np.zeros(layers[cat - 1].GetFeatureCount())
            for i, feature in enumerate(layers[cat - 1]):
                geom = feature.GetGeometryRef()
                geom.Transform(transform)  # To get area in m2
                areas[i] = geom.GetArea()
            # Delete largest polygon, which is outer border
            outer = np.where(areas == max(areas))[0]
            areas[outer] = 0
            total_area[cat - 1] = np.sum(areas)
            layers[cat - 1].DeleteFeature(outer)
            layers[cat - 1].ResetReading()

        # Calculate how many elements to be seeded for each category
        areas_weighted = total_area * thickness_microns
        numbers = number * areas_weighted / np.sum(areas_weighted)
        numbers = np.round(numbers).astype(int)
        oil_density = 1000
        mass_oil = (total_area * thickness_microns / 1e6) * oil_density

        for i, num in enumerate(numbers):
            self.seed_from_shapefile([mem_vector_layers[i]],
                                     oil_film_thickness=thickness_microns[i] /
                                     1000000.,
                                     mass_oil=mass_oil[i] / num,
                                     number=num,
                                     time=time,
                                     *args,
                                     **kwargs)

    def _substance_name(self):
        return self.get_oil_name()



# This file is part of OpenDrift.
# 
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
# 
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
# 
# Copyright 2015, Knut-Frode Dagestad, MET Norway

from opendrift.elements import LagrangianArray


class PassiveTracer(LagrangianArray):
    """Basic implementation of LagrangianArray with no additional properties.

    Contains only the properties of the abstract class LagrangianArray,
    i.e. position (lon, lat, z) and ID.
    May be used for passive tracer calculations when no properties are needed.
    """

    pass



'''
    Classes for storing measured values within an Oil record
'''
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json, JSON_List
from ..common.measurement import (Time,
                                  Temperature,
                                  MassFraction,
                                  Unitless,
                                  Dimensionless,
                                  Density,
                                  DynamicViscosity,
                                  KinematicViscosity,
                                  InterfacialTension,
                                  Pressure,
                                  AngularVelocity)

# from ..common.validators import EnumValidator
# from .validation.warnings import WARNINGS
# from .validation.errors import ERRORS


@dataclass_to_json
@dataclass
class InterfacialTensionPoint:
    tension: InterfacialTension
    ref_temp: Temperature
    interface: str = None
    method: str = None


class InterfacialTensionList(JSON_List):
    item_type = InterfacialTensionPoint


@dataclass_to_json
@dataclass
class Dispersibility:
    dispersant: str = None
    effectiveness: MassFraction = None
    method: str = None


class DispersibilityList(JSON_List):
    item_type = Dispersibility


@dataclass_to_json
@dataclass
class Emulsion:
    """
    There is no use for an empty Emulsion object

    But there is a use for one that has an arbitrary subset of fields,

    so making them all optional
    """
    age: Time = None
    water_content: MassFraction = None

    ref_temp: Temperature = None

    # Pa units, some kind of pressure/stress.
    # Adhesion provides the right units
    # but we shouldn't use it - so Pressure?
    complex_modulus: Pressure = None
    storage_modulus: Pressure = None
    loss_modulus: Pressure = None

    tan_delta_v_e: Unitless = None

    complex_viscosity: DynamicViscosity = None

    method: str = None
    visual_stability: str = None


class EmulsionList(JSON_List):
    item_type = Emulsion


@dataclass_to_json
@dataclass
class ESTSEvaporationTest:
    a_for_ev_a_b_ln_t_c: float = None
    a_for_ev_a_b_ln_t: float = None
    a_for_ev_a_b_sqrt_t: float = None
    b_for_ev_a_b_ln_t_c: float = None
    b_for_ev_a_b_ln_t: float = None
    b_for_ev_a_b_sqrt_t: float = None
    c_for_ev_a_b_ln_t_c: float = None
    method: str = None



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

# This ShipDrift model is based on
# Soergaard, E and T Vada (1998);
# Observations and modelling of drifting ships. Report DnV 96-2011
# Reprogrammed in Python for OpenDrift by Knut-Frode Dagestad Dec 2016

import os
import numpy as np
import scipy
import logging; logger = logging.getLogger(__name__)

from opendrift.models.basemodel import OpenDriftSimulation
from opendrift.elements import LagrangianArray


class ShipObject(LagrangianArray):
    """Extending LagrangianArray with variables relevant for leeway objects.

    """
    variables = LagrangianArray.add_variables([
        ('orientation', {'dtype': np.uint8,
                         'units': '1',
                         'default': 1}),
        ('length', {'dtype': np.float32,
                         'units': 'm',
                         'min': 1,
                         'max': 500,
            'description': 'Length of ship',
            'level': OpenDriftSimulation.CONFIG_LEVEL_ESSENTIAL,
                         'default': 80}),
        ('height', {'dtype': np.float32,
                         'units': 'm',
                         'min': 1,
                         'max': 100,
            'description': 'Total height of ship',
            'level': OpenDriftSimulation.CONFIG_LEVEL_ESSENTIAL,
                         'default': 8}),
        ('draft', {'dtype': np.float32,  # wet part of ship [m]
                         'units': 'm',
                         'min': 1,
                         'max': 30,
            'description': 'Draft of ship (depth below water)',
            'level': OpenDriftSimulation.CONFIG_LEVEL_ESSENTIAL,
                         'default': 4.0}),
        ('beam', {'dtype': np.float32,  # width of ship
                         'min': 1,
                         'max': 70,
                         'units': 'm',
            'description': 'Beam (width) of ship',
            'level': OpenDriftSimulation.CONFIG_LEVEL_ESSENTIAL,
                         'default': 10}),
        ('wind_drag_coeff', {'dtype': np.float32,  # Cf
                             'units': '1',
                             'default': 1}),
        ('water_drag_coeff', {'dtype': np.float32,  # Cd
                              'units': '1',
                              'default': 1}),
        ('jibeProbability', {'dtype': np.float32,
                             'units': '1/h',
                             'default': 0.04}),
       ])


class ShipDrift(OpenDriftSimulation):
    """Demonstration trajectory model based on OpenDrift framework.

    Simply advects a particle (passive tracer with
    no properties except for position) with the ambient wind.
    """

    ElementType = ShipObject

    required_variables = {
        'x_wind': {'fallback': None},
        'y_wind': {'fallback': None},
        'land_binary_mask': {'fallback': None},
        'x_sea_water_velocity': {'fallback': None},
        'y_sea_water_velocity': {'fallback': None},
        'sea_surface_wave_stokes_drift_x_velocity': {'fallback': 0},
        'sea_surface_wave_stokes_drift_y_velocity': {'fallback': 0},
        'sea_surface_wave_significant_height': {'fallback': 0},
        'sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment': {'fallback': 0}
        }

    max_speed = 2  # m/s
    winwav_angle = 20  # Angular offset in degrees

    def __init__(self, *args, **kwargs):

        # Read ship properties
        d = os.path.dirname(os.path.realpath(__file__))
        w = open(d + '/wforce.dat', 'r')
        self.wforce = {}
        w.readline()
        nbeam = int(w.readline().split()[0])
        self.wforce['nbeam'] = nbeam
        self.wforce['BL'] = np.array(w.readline().split()[0:nbeam],
                                     dtype=float)
        ndraft = int(w.readline().split()[0])
        self.wforce['ndraft'] = ndraft
        self.wforce['DL'] = np.array(w.readline().split()[0:ndraft],
                                     dtype=float)
        nomega = int(w.readline().split()[0])
        self.wforce['nomega'] = nomega
        self.wforce['omega'] = np.zeros((nomega))
        self.wforce['F'] = np.zeros((nomega, nbeam, ndraft))
        self.wforce['D'] = np.zeros((nomega, nbeam, ndraft))
        for o in range(nomega):
            self.wforce['omega'][o] = float(w.readline().split()[0])
            for i in range(ndraft):
                l = w.readline()
                self.wforce['F'][o, i, :] = l.split()[0:nbeam]
            for i in range(ndraft):
                l = w.readline()
                self.wforce['D'][o, i, :] = l.split()[0:nbeam]

        w.close()
        wi_omega, wi_BL, wi_DL = \
            np.meshgrid(self.wforce['omega'],
                        self.wforce['BL'], self.wforce['DL'], indexing='ij')
        self.wforce_interpolator_F = scipy.interpolate.LinearNDInterpolator(
            (wi_omega.ravel(), wi_BL.ravel(), wi_DL.ravel()),
             self.wforce['F'].ravel())
        self.wforce_interpolator_D = scipy.interpolate.LinearNDInterpolator(
            (wi_omega.ravel(), wi_BL.ravel(), wi_DL.ravel()),
             self.wforce['D'].ravel())

        super(ShipDrift, self).__init__(*args, **kwargs)

        self._add_config({'seed:orientation': {'type': 'enum',
            'enum':['left', 'right', 'random'], 'default': 'random',
            'level': OpenDriftSimulation.CONFIG_LEVEL_ESSENTIAL,
            'description': 'If ships are oriented to the left or right of the downwind direction,'
                'or whether this is unknown. Left/right means that wind will hit ship from backboard/steerboard'}})

        # Since the ShipDrift model is deterministic for given ship size,
        # (in contrast to the Leeway model), we use a default diffusivity
        # to yield some variability.
        self._set_config_default('drift:horizontal_diffusivity', 100)

    def seed_elements(self, *args, **kwargs):

        if 'number' in kwargs:
            num = kwargs['number']
        else:
            num = self.get_config('seed:number')
        for var in ['length', 'height', 'draft', 'beam']:
            if var not in kwargs:
                kwargs[var] = self.get_config('seed:' + var)
            kwargs[var] = np.atleast_1d(kwargs[var])
            if len(kwargs[var] == 1):
                kwargs[var] = kwargs[var]*np.ones(num)

        # Check that beam and height vs length are within expected range
        dl = kwargs['draft']/kwargs['length']
        if dl.min() < 0.025 or dl.max() > 0.07:
            logger.warning('Ratio of draft to length should be in range '
                                '0.025 to 0.07, given range is %s-%s. '
                                'Using border value.' %
                                (dl.min(), dl.max()))
            dl = np.clip(dl, 0.025, 0.07)
        bl = kwargs['beam']/kwargs['length']
        if bl.min() < 0.12 or bl.max() > 0.18:
            logger.warning('Ratio of beam to length should be in range '
                                '0.12 to 0.18, given range is %s-%s. '
                                'Using border value.' %
                                (bl.min(), bl.max()))

        # Calculate drag coefficients based on given ship dimensions
        # Wind drag coefficient
        exposed = kwargs['height'] - kwargs['draft']
        Cf = np.zeros(num)
        Cf[exposed>37.2] = 1.4
        Cf[exposed<=37.2] = 1.045 + 0.016*(exposed[exposed<=37.2] - 15.)
        Cf[exposed<=15] = 0.700 + 0.023*exposed[exposed<=15]
        kwargs['wind_drag_coeff'] = Cf

        # Water drag coefficient
        beta = 2.0*dl
        Cd = np.zeros(num)
        Cd[beta>.12] = 1.27
        Cd[beta<=.12] =  1.32 + (1.27-1.32)/0.02 * (beta[beta<=.12]-0.10)
        Cd[beta<=.10] =  1.38 + (1.32-1.38)/0.02 * (beta[beta<=.10]-0.08)
        Cd[beta<=.08] =  1.44 + (1.38-1.44)/0.02 * (beta[beta<=.08]-0.06)
        Cd[beta<=.06] =  1.50 + (1.44-1.50)/0.01 * (beta[beta<=.06]-0.05)
        kwargs['water_drag_coeff'] = Cd

        if 'orientation' not in kwargs:
            oc = self.get_config('seed:orientation')
            if oc == 'left':
                kwargs['orientation'] = np.ones(num)*0
            elif oc == 'right':
                kwargs['orientation'] = np.ones(num)*1
            else:
                kwargs['orientation'] = np.r_[:num] % 2  # Random 0 or 1

        # Calling general constructor with calculated values
        super(ShipDrift, self).seed_elements(*args, **kwargs)

    def update(self):

        Tm = self.wave_period()
        Hs = self.significant_wave_height()
        dl = self.elements.draft/self.elements.length
        bl = self.elements.beam/self.elements.length
        # Clip to allowed range
        bl = np.clip(bl, 0.12, 0.18)
        dl = np.clip(dl, 0.025, 0.07)
        # Additional clipping to avoid NaN from interpolator
        bl = np.clip(bl, 0.121, 0.179)
        dl = np.clip(dl, 0.0251, 0.069)

        # Simply move particles with ambient current
        self.update_positions(self.environment.x_sea_water_velocity,
                              self.environment.y_sea_water_velocity)

        # Wind force
        rho_air = 1.25  # to be checked
        area_dry = self.elements.length*(self.elements.height -
                                         self.elements.draft)
        area_wet = self.elements.length*self.elements.draft
        # Wind force
        F_wind = (0.5*rho_air*self.elements.wind_drag_coeff*
                  area_dry*np.power(self.wind_speed(), 2))
        # Decompose wind
        F_wind_x = F_wind*self.environment.x_wind/self.wind_speed()
        F_wind_y = F_wind*self.environment.y_wind/self.wind_speed()
        F_wind_x[self.wind_speed()==0] = 0
        F_wind_y[self.wind_speed()==0] = 0

        # Wave force
        rho_water = 1025
        NSPEC = 100
        ommin2 = 2.25
        ommin3 = 7.0
        ommax = 12.0
        dom = (ommax-ommin2)/(NSPEC-1)
        F_wave = np.zeros(self.num_elements_active())  # For each ship
        beta2 = 0
        scale1 = np.sqrt(9.81/self.elements.length)

        tmp = np.power(2.0*np.pi/Tm, 4)
        d = tmp*Hs*Hs/(4*np.pi)
        b = tmp/np.pi
        s = np.zeros((NSPEC, self.num_elements_active()))
        for i in range(NSPEC):
            omi = ommin2 + i*dom
            omi = omi*scale1
            s[i, :] = d * np.exp(-b/np.power(omi, 4)) / np.power(omi, 5)  # m2s

        f2 = 0.0
        d2 = 0.0
        for i in range(NSPEC):
            omi = ommin2 + i*dom
            f1 = f2
            d1 = d2
            #print omi - ommin3, 'should be positive'
            #print omi
            if (omi < ommin3):
                f2 = self.wforce_interpolator_F(omi, bl, dl)
                d2 = self.wforce_interpolator_D(omi, bl, dl)
            else:
                # Interval 3
                f2 = 0.5
                d2 = 4.0*omi*f2

            F_wave = F_wave + 0.5*(f1+f2)*dom*scale1*np.power(s[i,:], 2)
            beta2 = beta2 + 0.5*(d1+d2)*dom*scale1*np.power(s[i,:], 2)

        F_wave = F_wave*rho_water*9.81*self.elements.length
        beta2 = beta2*rho_water*np.sqrt(9.81*self.elements.length)

        # Add calculated wave and wind drift
        longperiod = Tm > 8.55
        F_wave_b = F_wave.copy()
        F_wave[longperiod] = F_wave[longperiod]*.66
        beta2[longperiod] = beta2[longperiod]*.60
        medperiod = ((Tm >= 5.7) & (Tm <= 8.55))
        F_wave[medperiod] = F_wave[medperiod]*(1.0 - 0.34*(Tm[medperiod]-5.7)/2.85)
        beta2[medperiod] = beta2[medperiod]*(1.0 - 0.4*(Tm[medperiod]-5.7)/2.85)

        # Form drag (water resistance)
        beta1 = 0.5*rho_water*self.elements.water_drag_coeff*area_wet

        # Wave direction is taken as wind direction plus offset +/- 20 degrees
        # Using minus, since angles are here defined counter-clockwise from x-axis
        offset = -self.winwav_angle*2*(self.elements.orientation - 0.5)
        if (self.environment.sea_surface_wave_stokes_drift_x_velocity.max() == 0 and
            self.environment.sea_surface_wave_stokes_drift_y_velocity.max() == 0):
                logger.info('Using wind direction as wave direction')
                wave_dir = np.radians(offset) + np.arctan2(self.environment.y_wind,
                                                           self.environment.x_wind)
        else:
            logger.info('Using Stokes drift direction as wave direction')
            wave_dir = np.radians(offset) + np.arctan2(
                self.environment.sea_surface_wave_stokes_drift_y_velocity,
                self.environment.sea_surface_wave_stokes_drift_x_velocity)
        F_wave_x = F_wave*np.cos(wave_dir)
        F_wave_y = F_wave*np.sin(wave_dir)
        F_total = np.sqrt(np.power(F_wind_x + F_wave_x, 2) +
                          np.power(F_wind_y + F_wave_y, 2))

        # Iterate 4 times in order to estimate the effect of
        # wave damping and form drag
        uw_tot = 0
        uw_dir = 0
        for i in range(4):
            # Calc wave damping force in x and y (Ref (1), eq. 6.6.4)
            f2x = beta2*uw_tot*np.cos(wave_dir)
            f2y = beta2*uw_tot*np.sin(wave_dir)
            # Calc new drift direction, including effect of
            # wave damping (Ref (1), eq. 6.6.1)
            uw_dir = np.arctan2(F_wind_y+F_wave_y-f2y, F_wind_x+F_wave_x-f2x)
            # Ref (1), eq. 6.6.3
            bet2c = beta2*np.cos((wave_dir-uw_dir))
            # Ref (1), eq. 6.4.3
            uw_tot = -bet2c/(2.*beta1) + np.sqrt(bet2c*bet2c +
                                                 4.*beta1*F_total)/(2.*beta1)

        # Finally advect according to wind-wave forces
        velocity_u = uw_tot*np.cos(uw_dir)
        velocity_v = uw_tot*np.sin(uw_dir)
        self.update_positions(velocity_u, velocity_v)

        # Stranding
        self.deactivate_elements(self.environment.land_binary_mask == 1,
                                 reason='ship stranded')

    def import_ascii_format(self, filename):

        with open(filename) as f:
            lines = f.readlines()
        self.time_step_output, self.time_step = lines[16].split()
        num_elements = int(lines[25].split()[0])
        if num_elements != 1:
            raise ValueError('Import presently only supports single ship')
        num_timesteps = (len(lines)-30.)/14.
        num_timesteps = int(num_timesteps)

        # Initialise history array
        from datetime import datetime, timedelta
        history_dtype_fields = [
            (name, self.ElementType.variables[name]['dtype'])
            for name in self.ElementType.variables]
        # Add environment variables
        self.history_metadata = self.ElementType.variables.copy()
        for env_var in self.required_variables:
            history_dtype_fields.append((env_var, np.dtype('float32')))
            self.history_metadata[env_var] = {}
        history_dtype = np.dtype(history_dtype_fields)

        self.history = np.ma.array(
	        np.zeros([num_elements, num_timesteps]),
            dtype=history_dtype, mask=[True])

        self.steps_output = num_timesteps
        self.steps = num_timesteps
        self.start_time = datetime.now()
        self.end_time = datetime.now() + timedelta(hours=1)
        self.time = self.end_time

        # Read time steps from file
        for i in range(num_timesteps):
            line = lines[30 + 14*i + 13]
            l = line.split()
            lon = float(l[2])
            lat = float(l[3])
            self.history['lon'][0, i] = lon
            self.history['lat'][0, i] = lat



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

from opendrift.readers.basereader import BaseReader, ContinuousReader
import numpy as np


class Reader(BaseReader, ContinuousReader):
    '''A very simple reader that always give the same value for its variables'''

    def __init__(self, parameter_value_map):
        '''init with a map {'variable_name': value, ...}'''

        for key, var in parameter_value_map.items():
            parameter_value_map[key] = np.atleast_1d(var)
        self._parameter_value_map = parameter_value_map
        self.variables = list(parameter_value_map.keys())
        self.proj4 = '+proj=latlong'
        self.xmin = -180
        self.xmax = 180
        self.ymin = -90
        self.ymax = 90
        self.start_time = None
        self.end_time = None
        self.time_step = None
        self.name = 'constant_reader'

        # Run constructor of parent Reader class
        super(Reader, self).__init__()

    def get_variables(self, requestedVariables, time=None,
                      x=None, y=None, z=None):

        variables = {'time': time, 'x': x, 'y': y, 'z': z}
        #variables.update(self._parameter_value_map)
        for var in requestedVariables:
            variables[var] = self._parameter_value_map[var]*np.ones(x.shape)

        return variables




# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

import logging; logging.captureWarnings(True); logger = logging.getLogger(__name__)
from opendrift.readers.basereader import BaseReader
from opendrift.readers import reader_from_url


class Reader:
    '''For lazy initialisation'''

    logger = logging.getLogger('opendrift')  # using common logger

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self.initialised = False
        self._lazyname = 'LazyReader: ' + args[0]
        logger.debug('Delaying initialisation of ' + self._lazyname)

    def __getattr__(self, name):
        if self.initialised is False:
            if name == 'name':
                return self._lazyname
            if name == 'is_lazy':
                return True
            self.initialise()

        try:
            return object.__getattribute__(self.reader, name)
        except:
            return object.__getattribute__(self, name)

    def get_variables(self, *args, **kwargs):
        return self.reader.get_variables(*args, **kwargs)

    def initialise(self):
        logger.debug('Initialising: ' + self._lazyname)
        self.reader = reader_from_url(self._args[0])
        if self.reader is None:
            raise ValueError('Reader could not be initialised')
        else:
            if 'prepare_args' in dir(self):
                self.reader.prepare(**self.prepare_args)
            logger.debug('Reader initialised: ' + self.reader.name)
            self.initialised = True

    def prepare(self, **kwargs):
        self.prepare_args = kwargs

    def __repr__(self):
        if self.initialised is True:
            return self.reader.__repr__()
        else:
            return 'Lazy reader:\n%s\n%s' % (
                        self._args, self._kwargs)



#
# Model class definitions for embedded content in our oil records
#
# Note: These are deprecated a bit.  An updated SARAFraction is sitting in
#       models.oil
from dataclasses import dataclass
from ..common.utilities import dataclass_to_json

from .enum_types import SaraTypeEnum


@dataclass_to_json
@dataclass
class SARAFraction:
    sara_type: SaraTypeEnum

    fraction: float
    ref_temp_k: float = 273.15
    weathering: float = 0.0

    standard_deviation: float = None
    replicates: float = None
    method: str = None

    def __post_init__(self):
        self.sara_type = SaraTypeEnum(self.sara_type)
        self.fraction = float(self.fraction)
        self.ref_temp_k = float(self.ref_temp_k)
        self.weathering = float(self.weathering)

        if self.standard_deviation is not None:
            self.standard_deviation = float(self.standard_deviation)

        if self.replicates is not None:
            self.replicates = float(self.replicates)

        if self.method is not None:
            self.method = str(self.method)


@dataclass_to_json
@dataclass
class SARADensity:
    sara_type: SaraTypeEnum

    kg_m_3: float
    ref_temp_k: float = 273.15
    weathering: float = 0.0

    def __post_init__(self):
        self.sara_type = SaraTypeEnum(self.sara_type)
        self.kg_m_3 = float(self.kg_m_3)
        self.ref_temp_k = float(self.ref_temp_k)
        self.weathering = float(self.weathering)


@dataclass_to_json
@dataclass
class MolecularWeight:
    sara_type: SaraTypeEnum

    g_mol: float
    ref_temp_k: float = 273.15
    weathering: float = 0.0

    def __post_init__(self):
        self.sara_type = SaraTypeEnum(self.sara_type)
        self.g_mol = float(self.g_mol)
        self.ref_temp_k = float(self.ref_temp_k)
        self.weathering = float(self.weathering)

    @property
    def kg_mol(self):
        return self.g_mol / 1000.0

    @kg_mol.setter
    def length(self, value):
        self.g_mol = value * 1000.0



from abc import abstractmethod
from numbers import Number
from typing import List

class Combine:
    def __add__(self, other):
        from .readerops import Combined as ReaderCombined
        from .numops import Combined as NumCombined
        from ..basereader import BaseReader

        if isinstance(other, Number):
            return NumCombined.add(other, self)
        elif isinstance(other, BaseReader):
            return ReaderCombined(self, other, lambda a, b: a + b)
        else:
            return NotImplemented

    def __mul__(self, other):
        from .numops import Combined as NumCombined
        if isinstance(other, Number):
            return NumCombined.mul(other, self)
        else:
            return NotImplemented

    def __truediv__(self, other):
        from .numops import Combined as NumCombined

        if isinstance(other, Number):
            return NumCombined.div(other, self)
        else:
            return NotImplemented

    def __sub__(self, other):
        return self + (-1 * other)

class Filter:
    @property
    @abstractmethod
    def variables(self) -> List[str]:
        pass

    def filter_vars(self, vars):
        """
        Only keep the specified variables.
        """
        from .filter import FilterVariables
        return FilterVariables(self, vars)

    def exclude_vars(self, vars):
        """
        Remove the specified variables.
        """
        from .filter import FilterVariables

        vars = list(set(self.variables) - set(vars))
        return FilterVariables(self, vars)




'''
    class that represents the demographic data (metadata) of an oil record.
'''
from datetime import datetime
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json
from ..common.measurement import MassFraction, Temperature

from .values import Reference
from .product_type import ProductType, DOESNT_NEED_API
from .location_coordinates import LocationCoordinates

from .validation.warnings import WARNINGS
from .validation.errors import ERRORS


@dataclass_to_json
@dataclass
class MetaData:
    name: str = ''
    source_id: str = ''
    alternate_names: list = field(default_factory=list)
    location: str = ''
    reference: Reference = field(default_factory=Reference)
    sample_date: str = ''
    product_type: ProductType = ''
    API: float = None
    comments: str = ''
    labels: list = field(default_factory=list)
    model_completeness: float = None
    location_coordinates: LocationCoordinates = None
    gnome_suitable: bool = None

    def validate(self):
        msgs = []
        # check for API
        api = self.API
        if api is None:
            if self.product_type in DOESNT_NEED_API:
                msgs.append(WARNINGS["W004"])
            else:
                msgs.append(ERRORS["E030"])
        else:
            if not (-60.0 < api < 100):  # somewhat arbitrary limits
                msgs.append(WARNINGS["W005"].format(api=api))

        # Check for a reasonable name
        # right now, reasonable is more than 5 characters -- we may want to add more later
        if len(self.name.strip()) < 2:
            msgs.append(WARNINGS["W001"].format(self.name))

        # check sample date is valid
        if self.sample_date:
            try:
                datetime.fromisoformat(self.sample_date)
            except ValueError as err:
                msgs.append(WARNINGS["W011"].format("sample date", self.sample_date, str(err)))
        return msgs


@dataclass_to_json
@dataclass
class SampleMetaData:
    name: str = "Fresh Oil Sample"
    short_name: str = None
    sample_id: str = None
    description: str = None
    fraction_evaporated: MassFraction = None
    boiling_point_range: Temperature = None



import cartopy.crs as ccrs
import pyproj
import utm


def simulation():
    """
    Simulation Spatial Reference System (SRS) or Coordinate Reference System (CRS).

    Always just 'lonlat' ellipsoid 'WGS84'
    """
    return pyproj.Proj('+proj=lonlat')


def find_utm_proj(lon, lat):
    """
    Find a suitable UTM projection (zone) for lon and lat.

    .. warning::

      UTM is only defined between 80S and 84N. Should use UPS for those regions.

    Returns:

      pyproj.Proj in `utm` projection.
    """
    _, _, zone_no, _ = utm.from_latlon(lat, lon)
    band = 'south' if lat < 0 else 'north'

    return pyproj.Proj(
        '+proj=utm +zone={zone:d} +{band} +ellps=WGS84 +datum=WGS84 +units=m +no_defs'
        .format(zone=zone_no, band=band))

def find_utm_ccrs(lon, lat):
  """
  Return Cartopy UTM CRS for lon and lat.
  """
  _, _, zone_no, _ = utm.from_latlon(lat, lon)
  return ccrs.UTM(zone_no, True if lat < 0 else False)




'''
    Oil Properties Estimations.

    These are primitive estimation algorithms to be used primarily for
    estimating oil properties based on known measured values.
'''

import numpy as np


# fixme -- these are unit conversion, and are in the unit_conversion(NUCOS)
#          library.  We should use that lib to be consistent.

def density_from_api(api):
    '''
        Source: Adios2

        API is a measure of how heavy an oil is compared to water.
        So it is a different unit for specific gravity
    '''
    kg_m_3 = 141.5 / (131.5 + api) * 1000.0
    ref_temp_k = 273.15 + 15.0

    return kg_m_3, ref_temp_k


def api_from_density(density):
    '''
        Source: Adios2
    '''
    kg_m_3 = density

    return (141.5 / kg_m_3 * 1000.0) - 131.5


def density_at_temp(ref_density, ref_temp_k, temp_k, k_rho_t=0.0008):
    '''
        Source: Adios2

        If we have an oil density at a reference temperature,
        then we can estimate what its density might be at
        another temperature.

        NOTE: need a reference for the coefficient of expansion
    '''
    return ref_density / (1.0 - k_rho_t * (ref_temp_k - temp_k))


def vol_expansion_coeff(rho_0, t_0, rho_1, t_1):
    '''
        Calculate the volumetric expansion coefficient of a liquid
        based on a set of two densities and their associated temperatures.
    '''
    if t_0 == t_1:
        k_rho_t = 0.0
    else:
        k_rho_t = (rho_0 - rho_1) / (rho_0 * (t_1 - t_0))

    return k_rho_t


def specific_gravity(density):
    '''
        Specific Gravity of Oil with respect to water at 15C (definition used
        for API gravity)
    '''
    return density / 999.103


def dvis_to_kvis(dvis, density):
    '''
        Source: Definition of kinematic viscosity.

        Conversion from dynamic viscosity to kinematic viscosity.
    '''
    return dvis / density


def kvis_at_temp(ref_kvis, ref_temp_k, temp_k, k_v2=2416.0):
    '''
        Source: Adios2

        If we have an oil kinematic viscosity at a reference temperature,
        then we can estimate what its viscosity might be at
        another temperature.

        Note: Bill's most recent viscosity document, and an analysis of the
              multi-KVis oils in our oil library suggest that a value of
              2416.0 (Abu Eishah 1999) would be a good default value for k_v2.
    '''
    return ref_kvis * np.exp(k_v2 / temp_k - k_v2 / ref_temp_k)


def resin_fraction(density, viscosity, f_other=0.0):
    A = _A_coeff(density)
    B = _B_coeff(density, viscosity)

    f_res = 0.033 * A + 0.00087 * B - 0.74
    f_res = np.clip(f_res, 0.0, 1.0 - f_other)

    return f_res


def asphaltene_fraction(density, viscosity, f_other=0.0):
    A = _A_coeff(density)
    B = _B_coeff(density, viscosity)

    f_asph = (0.000014 * A ** 3.0 +
              0.000004 * B ** 2.0 -
              0.18)
    f_asph = np.clip(f_asph, 0.0, 1.0 - f_other)

    return f_asph


def saturates_fraction(density, viscosity, f_other=0.0):
    A = _A_coeff(density)
    B = _B_coeff(density, viscosity)

    f_sat = -2.5 + 76.6 / A + 0.00013 * np.log(B)
    f_sat = np.clip(f_sat, 0.0, 1.0 - f_other)

    return f_sat


def aromatics_fraction(f_res, f_asph, f_sat):
    f_arom = 1.0 - (f_res + f_asph + f_sat)
    f_arom = np.clip(f_arom, 0.0, 1.0)

    return f_arom


def _A_coeff(density):
    '''
        Source: Fingas empirical formulas that are based upon analysis
                of ESTC oil properties database.

        This is an intermediate calculation for a coefficient to be
        used to generate the inert mass fractions of an oil.
    '''
    return 10.0 * np.exp(0.001 * density)


def _B_coeff(density, viscosity):
    '''
        Source: Fingas empirical formulas that are based upon analysis
                of ESTC oil properties database.

        This is an intermediate calculation for a coefficient to be
        used to generate the inert mass fractions of an oil.
    '''
    return 10.0 * np.log(1000.0 * density * viscosity)


def cut_temps_from_api(api, N=5):
    '''
        Source: Adios2 & Jones R. (1997),
                A Simplified Pseudo-component Oil Evaporation Model,
                Proceedings of the 20th Arctic and Marine Oil Spill Program,
                Vancouver, CA,
                Vol. 1, pp. 43-62

        Generate distillation cut temperatures from the oil's API.
    '''
    T_0 = 457.0 - 3.34 * api
    T_G = 1357.0 - 247.7 * np.log(api)

    return np.array([(T_0 + T_G * i / N) for i in range(N)])


def fmasses_from_cuts(f_evap_i):
    '''
        Generate distillation cut fractional masses from the
        cumulative distillation fractions in the cut data.
    '''
    fmass_i = np.array(f_evap_i)
    fmass_i[1:] = np.diff(fmass_i)

    return fmass_i


def fmasses_flat_dist(f_res, f_asph, N=5):
    '''
        Generate a flat distribution of N distillation cut fractional masses.
    '''
    return np.array([(1.0 - f_res - f_asph) / N] * N)


def saturate_mol_wt(boiling_point):
    '''
        Source: Dr. M. R. Riazi,
                Characterization and Properties of Petroleum Fractions
                eq. 2.48 and table 2.6

        Note: for this to actually work in every case, we need to limit
              our temperature to:
              - T_i < 1070.0
              - T_i >
              - T_i > 1070.0 - exp(6.98291)  (roughly about == -8.06)
    '''
    T_i = np.clip(np.array(boiling_point),
                  1070.0 - np.exp(6.98291) + 0.00001,
                  1070.0 - 0.00001)
    return (49.677 * (6.98291 - np.log(1070.0 - T_i))) ** (3.0 / 2.0)


def aromatic_mol_wt(boiling_point):
    '''
        Source: Dr. M. R. Riazi,
                Characterization and Properties of Petroleum Fractions
                eq. 2.48 and table 2.6

        Note: for this to actually work in every case, we need to limit
              our temperature to:
              - T_i < 1015.0
              - T_i > 1015.0 - exp(6.911)  (roughly about == 11.76)
    '''
    T_i = np.clip(np.array(boiling_point),
                  1015.0 - np.exp(6.911) + 0.00001,
                  1015.0 - 0.00001)
    return (44.504 * (6.911 - np.log(1015.0 - T_i))) ** (3.0 / 2.0)


def resin_mol_wt(_boiling_point):
    '''
        Source: Recommendation from Bill Lehr

        Note: We pass in a boiling point to remain consistent with the other
              molecular weight functions, even though it is not used.
        Note: We return a scalar in all cases.  This should still work with
              numpy array operations, but probably not with regular Python
              sequence types.
              We can fix this if the need arises.
    '''
    return 800.0


def asphaltene_mol_wt(_boiling_point):
    '''
        Source: Recommendation from Bill Lehr

        Note: We pass in a boiling point to remain consistent with the other
              molecular weight functions, even though it is not used.
        Note: We return a scalar in all cases.  This should still work with
              numpy array operations, but probably not with regular Python
              sequence types.
              We can fix this if the need arises.
    '''
    return 1000.0


def trial_densities(boiling_points, watson_factor):
    '''
        Source: Dr. M. R. Riazi,
                Characterization and Properties of Petroleum Fractions
                eq. 2.13 and table 9.6

        Generate an initial estimate of volatile oil components based
        on boiling points and the Watson Characterization Factor.
        This is only good for estimating Aromatics & Saturates.
    '''
    boiling_points = np.array(boiling_points)
    return 1000.0 * (1.8 * boiling_points) ** (1.0 / 3.0) / watson_factor


def saturate_densities(boiling_points):
    K_w_sat = 12.0

    return trial_densities(boiling_points, K_w_sat)


def aromatic_densities(boiling_points):
    K_w_arom = 10.0

    return trial_densities(boiling_points, K_w_arom)


def resin_densities(_boiling_points):
    '''
        Note: We pass in a boiling point to remain consistent with the other
              molecular weight functions, even though it is not used.
        Note: We return a scalar in all cases.  This should still work with
              numpy array operations, but probably not with regular Python
              sequence types.
              We can fix this if the need arises.
    '''
    return 1100.0


def asphaltene_densities(_boiling_points):
    '''
        Note: We pass in a boiling point to remain consistent with the other
              molecular weight functions, even though it is not used.
        Note: We return a scalar in all cases.  This should still work with
              numpy array operations, but probably not with regular Python
              sequence types.
              We can fix this if the need arises.
    '''
    return 1100.0


def _hydrocarbon_characterization_param(specific_gravity, temp_k):
    '''
        Source: Dr. M. R. Riazi,
                Characterization and Properties of Petroleum Fractions
                eq. 2.115

        Source: Huang, P. K. (1977),
                Characterization and Thermodynamic Correlations for
                Undefined Hydrocarbon Mixtures,
                Ph.D. Dissertation
                Pennsylvania State University,
                University Park, PA,

        This is a characterization parameter, designated as I, that
        was first used by Huang to correlate hydrocarbon properties
    '''
    T_i = np.array(temp_k)
    SG_i = np.array(specific_gravity)

    return 0.3773 * T_i ** (-0.02269) * SG_i ** (0.9182)


def refractive_index(hc_char_param):
    '''
        Source: Dr. M. R. Riazi,
                Characterization and Properties of Petroleum Fractions
                eq. 2.114

        This is the refractive index of liquid hydrocarbons at 20C,
        correlated through parameter I
    '''
    i = np.array(hc_char_param)

    return ((1 + 2 * i) / (1 - i)) ** (1.0 / 2.0)


def _hydrocarbon_grouping_param(mol_wt, specific_gravity, temp_k):
    '''
        Source: Dr. M. R. Riazi,
                Characterization and Properties of Petroleum Fractions
                eq. 3.50

        This is a hydrocarbon grouping parameter correlated through
        the molecular weight and the refractive index.  Riazi claims
        that it:
        - separates paraffins and aromatics
        - identifies various hydrocarbon types
    '''
    i = _hydrocarbon_characterization_param(specific_gravity, temp_k)
    n = refractive_index(i)

    return mol_wt * (n - 1.475)


def saturate_mass_fraction(fmass_i, mol_wt, specific_gravity, temp_k):
    '''
        Source: Dr. M. R. Riazi,
                Characterization and Properties of Petroleum Fractions
                eqs. 3.77 and 3.78

        For a petroleum hydrocarbon fraction in which the molecular
        weight, specific gravity, and boiling point are known, we can
        estimate the fraction of that substance which contains
        saturate compounds.
        When forming his equations, Riazi deals with the PNA composition,
        which essentially means Paraffins, Naphthenes, and Aromatics.
        And for our purposes, the saturates include the paraffins and
        naphthenes.
    '''
    SG_sat_i = specific_gravity
    m = _hydrocarbon_grouping_param(mol_wt, SG_sat_i, temp_k)

    X_P = 3.7387 - 4.0829 * SG_sat_i + 0.014772 * m
    X_N = -1.5027 + 2.10152 * SG_sat_i - 0.02388 * m

    f_sat_i = fmass_i * (X_P + X_N)

    return np.clip(f_sat_i, 0.0, fmass_i)


def oil_water_surface_tension_from_api(api):
    return 0.001 * (39.0 - 0.2571 * api)


def pour_point_from_kvis(ref_kvis, ref_temp_k):
    '''
        Source: Adios2

        If we have an oil kinematic viscosity at a reference temperature,
        then we can estimate what its pour point might be.
    '''
    c_v1 = 5000.0
    ref_kvis = np.array(ref_kvis)
    ref_temp_k = np.array(ref_temp_k)

    T_pp = (c_v1 * ref_temp_k) / (c_v1 - ref_temp_k * np.log(ref_kvis))

    return T_pp


def pour_point_from_sg_mw_kvis(specific_gravity, mol_wt, kvis):
    '''
        Source: Dr. M. R. Riazi,
                Characterization and Properties of Petroleum Fractions
                eq. 3.119

        Another way of estimating pour point.
        These inputs may not be available for most imported oil records.
    '''
    SG = specific_gravity

    return (130.47 *
            SG ** 2.970566 *
            mol_wt ** (0.61235 - 0.47357 * SG) *
            kvis * (0.310331 - 0.32834 * SG))


def flash_point_from_bp(temp_k):
    '''
        Source: Reference: Chang A., K. Pashakanti, and Y. Liu (2012),
                           Integrated Process Modeling and Optimization,
                           Wiley Verlag.

    '''
    temp_k = np.array(temp_k)
    return 117.0 + 0.69 * temp_k


def flash_point_from_api(api):
    '''
        Source: Reference: Chang A., K. Pashakanti, and Y. Liu (2012),
                           Integrated Process Modeling and Optimization,
                           Wiley Verlag.

    '''
    api = np.array(api)
    return 457.0 - 3.34 * api


def bullwinkle_fraction_from_asph(f_asph):
    '''
        Source: Adios2
    '''
    return 0.32 - 3.59 * f_asph


def bullwinkle_fraction_from_api(api):
    '''
        Source: Adios2
    '''
    return 0.5762 * np.log10(api) - 0.6353



"""
utilities for doing computation on the physical properties of an
oil record
"""

from operator import itemgetter
import numpy as np

import unit_conversion as uc


class Density:
    """
    class to hold and do calculations on density

    data is stored internally in standard units:
    temperature in Kelvin
    density in kg/m^3
    """
    def __init__(self, oil):
        """
        Initialize a density calculator

        :param oil: an Oil object -- the density data will be extracted

        or

        :param oil: Sequence of density/temperature pairs:
                    ``[(980.0, 288.15), (990.0, 273.15)])``

        If data pairs, units must be kg/m^3 and K
        """
        try:
            data = get_density_data(oil, units='kg/m^3', temp_units="K")
        except AttributeError:
            # not an oil object -- assume it's a table of data in the
            #                      correct form
            data = oil
        if data:
            data = sorted(data, key=itemgetter(1))
            self.densities, self.temps = zip(*data)
        else:
            self.densities = []
            self.temps = []
        self.initialize()

    def initialize(self):
        """
        Initialize the expansion coefficient

        For outside the measured range
        """
        # if there is only one density, use a default
        # Note: no idea where these values came from

        if not np.all(np.diff(self.temps) > 0):
            raise ValueError("temperatures must be discreet")

        if len(self.densities) == 1:
            d = self.densities[0]
            t = self.temps[0]
            if abs(t - 288.16) < 5.0:  # measurement within 5 deg of 15 C
                # API 30 threshold
                self.k_rho_default = -0.0009 if d < 875 else -0.0008
            else:
                self.k_rho_default = -0.00085  # who knows?
        elif len(self.densities) > 1:
            # do a linear fit to the points
            # this should exactly match if there are only two.
            b = self.densities
            A = np.c_[np.ones_like(b), np.array(self.temps)]
            x, _residuals, _rank, _s = np.linalg.lstsq(A, b, rcond=None)
            self.k_rho_default = x[1]
        else:
            raise ValueError("Density needs at least one density value")

    def at_temp(self, temp, unit='K'):
        """
        density(s) at the provided temperature(s)

        :param temp: scalar or sequence of temp in K

        :param unit='K': unit of temperature

        densities will be returned as kg/m^3
        """
        temp = np.asarray(temp)
        scaler = True if temp.shape == () else False
        temp.shape = (-1, )

        if unit != 'K':
            temp = uc.convert(unit, 'K', temp)

        densities = np.interp(temp, self.temps, self.densities,
                              left=-np.inf, right=np.inf)

        left = (densities == -np.inf)
        densities[left] = self.densities[0] + (self.k_rho_default *
                                               (temp[left] - self.temps[0]))

        right = (densities == np.inf)
        densities[right] = self.densities[-1] + (self.k_rho_default *
                                                 (temp[right] - self.temps[-1]))

        return densities if not scaler else densities[0]


class KinematicViscosity:
    """
    class to hold and do calculations on kinematic viscosity

    data is stored internally in standard units:
    temperature in Kelvin
    viscosity in m^2/s
    """
    def __init__(self, oil):
        """
        initialize from an oil object
        """
        try:
            data = get_kinematic_viscosity_data(oil, units='m^2/s',
                                                temp_units="K")
        except AttributeError:
            # not an oil object -- assume it's a table of data in the
            #                      correct form
            data = oil

        if data:
            data = sorted(data, key=itemgetter(1))
            self.kviscs, self.temps = zip(*data)
        else:
            self.kviscs = []
            self.temps = []
        self.initialize()

    def at_temp(self, temp, kvis_units='m^2/s', temp_units="K"):
        """
        Compute the kinematic viscosity of the oil as a function of temperature

        :param temp_k: temperatures to compute at: can be scalar or array
                       of values.  Should be in Kelvin

        viscosity as a function of temp is given by:
        v = A exp(k_v2 / T)

        with constants determined from measured data
        """
        temp = np.asarray(temp)
        temp = uc.convert('temperature', temp_units, 'K', temp)

        kvisc = self._visc_A * np.exp(self._k_v2 / temp)

        kvisc = uc.convert('kinematic viscosity', 'm^2/s', kvis_units, kvisc)
        return kvisc

    def initialize(self):
        '''
        viscosity as a function of temp is given by:

        v = A exp(k_v2 / T)

        The constants, A and k_v2 are determined from the viscosity data:

        If only one data point, a default value for k_vs is used:
           2100 K, based on analysis of data in the ADIOS database as of 2018

        If two data points, the two constants are directly computed

        If three or more, the constants are computed by a least squares fit.
        '''
        # find viscosity measurements with zero weathering

        # this sets:
        self._k_v2 = None  # decay constant for viscosity curve
        self._visc_A = None

        kvis = self.kviscs
        kvis_ref_temps = self.temps

        if len(kvis) == 1:  # use default k_v2
            self._k_v2 = 2100.0
            self._visc_A = kvis[0] * np.exp(-self._k_v2 / kvis_ref_temps[0])
        else:
            # do a least squares fit to the data
            # viscs = np.array(kvis)
            # temps = np.array(kvis_ref_temps)
            b = np.log(kvis)
            A = np.c_[np.ones_like(b), 1.0 / np.array(kvis_ref_temps)]
            x, _residuals, _rank, _s = np.linalg.lstsq(A, b, rcond=None)
            self._k_v2 = x[1]
            self._visc_A = np.exp(x[0])
        return


def get_density_data(oil, units="kg/m^3", temp_units="K"):
    """
    Return a table of density data:

    list of (density, temp) pairs

    :param oil: the oil object to get data from

    :param units="kg/m^3": units you want the density in

    :param temp_units="K": units you want the density in

    """

    densities = [d for d in oil.sub_samples[0].physical_properties.densities
                 if d.density is not None
                 and d.ref_temp is not None]

    # create normalized list of densities
    density_table = []
    for density_point in densities:
        d = density_point.density.converted_to(units).value
        t = density_point.ref_temp.converted_to(temp_units).value
        density_table.append((d, t))
    return density_table


def get_kinematic_viscosity_data(oil, units="m^2/s", temp_units="K"):
    """
    Return a table of kinematic viscosity data:

    list of (viscosity, temp) pairs

    :param oil: the oil object to get data from

    :param units="cSt": units you want the viscosity in

    :param temp_units="K": units you want the viscosity in

    """

    try:
        kvisc = [k for k
                 in oil.sub_samples[0].physical_properties.kinematic_viscosities
                 if k.viscosity is not None
                 and k.ref_temp is not None]
    except IndexError:  # no subsamples at all!
        return []

    if len(kvisc) > 0:
        visc_table = []
        for visc_point in kvisc:
            d = visc_point.viscosity.converted_to(units).value
            t = visc_point.ref_temp.converted_to(temp_units).value
            visc_table.append((d, t))
        return visc_table

    dvisc = oil.sub_samples[0].physical_properties.dynamic_viscosities
    if len(dvisc) > 0:
        dvisc = get_dynamic_viscosity_data(oil, units="Pa s", temp_units="K")
        density = Density(oil)
        visc_table = convert_dvisc_to_kvisc(dvisc, density)
        return visc_table
    else:
        return []


def get_dynamic_viscosity_data(oil, units="Pas", temp_units="K"):
    """
    Return a table of kinematic viscosity data:

    list of (viscosity, temp) pairs

    :param oil: the oil object to get data from

    :param units="cSt": units you want the viscosity in

    :param temp_units="K": units you want the viscosity in

    """

    dvisc = [d for d
             in oil.sub_samples[0].physical_properties.dynamic_viscosities
             if d.viscosity is not None
             and d.ref_temp is not None]

    if len(dvisc) > 0:
        visc_table = []
        for visc_point in dvisc:
            v = visc_point.viscosity.converted_to(units).value
            t = visc_point.ref_temp.converted_to(temp_units).value
            visc_table.append((v, t))
        return visc_table

    kvisc = oil.sub_samples[0].physical_properties.kinematic_viscosities
    if len(kvisc) > 0:
        raise NotImplementedError("can't compute dynamic from kinematic yet")
        kvisc = get_kinematic_viscosity_data(oil)
        # convert here.
    else:
        return []


def convert_dvisc_to_kvisc(dvisc, density):
    """
    convert dynamic viscosity to kinematic viscosity

    :param density: an initialized Density object

    dvisc and densities are tables as returned from:

     - ``get_dynamic_viscosity_data``
     - ``get_density_data``

    units: viscosity: Pas or kg/(m s)
           density: kg/m^3
    """
    kvisc_table = []
    for (dv, temp) in dvisc:
        kv = dv / density.at_temp(temp)
        kvisc_table.append((kv, temp))
    return kvisc_table


# def density_at_temp(densities, temp, units="kg/m^3", temp_units="K"):
#     # sort them to make sure
#     densities = sorted(densities, key=itemgetter(1))
#     dens, temps = zip(*densities)

#     return np.interp(temp, temps, dens)

# def get_kinematic_viscosity_at_temp(temp,
#                                     kvis_units='cSt',
#                                     temp_units='C'):
#     raise NotImplementedError


def get_pour_point(oil):
    """
    Return oil's pour point or None
    """
    phys_props = oil.sub_samples[0].physical_properties

    pour_point = phys_props.pour_point

    return pour_point


def get_distillation_cuts(oil, units="fraction", temp_units="K"):
    """
    Return a table of distillation data:

    list of (cut fraction, temp) pairs

    :param oil: the oil object to get data from

    :param units="fraction": units you want the fraction in

    :param temp_units="K": units you want the temperature in

    """

    distillation_cuts = oil.sub_samples[0].distillation_data.cuts

    # create normalized list of densities
    cuts_table = []
    for cut in distillation_cuts:
        f = cut.fraction.converted_to(units).value
        t = cut.vapor_temp.converted_to(temp_units).value
        cuts_table.append((f, t))
    cuts_table.sort(key=itemgetter(0))
    return cuts_table


def max_water_fraction_emulsion(oil):
    if (oil.metadata.product_type == 'Crude Oil NOS' or
            oil.metadata.product_type == 'Bitumen Blend'):
        return 0.9
    else:
        return 0.0


def emul_water(oil):
    """
    This function computes two terms used in emulsification.
    Ymax is the maximum water fraction of a stable emulsion.
    Smax is the maximum surface area of the water droplets inside
    the emulsion. (from ADIOS2)
    """

    emulsion_max_water = None  # need to drill down in database for this info

    if emulsion_max_water is None:  # max water content not in library
        dens = Density(oil)
        density = dens.at_temp(288.15)
        kvis = KinematicViscosity(oil)
        viscosity = kvis.at_temp(288.15)
        dynamic_viscosity = viscosity * density
        if (dynamic_viscosity > 0.050):
            Ymax = 0.9 - 0.0952 * np.log(dynamic_viscosity / 0.050)
        else:
            Ymax = 0.9
    else:
        Ymax = emulsion_max_water  # stay with the value you have

    # this is done is py_gnome
    # drop_min = 1.0e-6		# min oil droplet size
    # Smax = (6.0 / drop_min) * (Ymax / (1.0 - Ymax))
    return Ymax


def bullwinkle_fraction(oil):

    Ni = 0
    Va = 0
    try:
        bulk_composition = oil.sub_samples[0].bulk_composition
        for element in bulk_composition:
            if element.name == "nickel":
                Ni = element.measurement.value
    except Exception:
        Ni = 0.

    try:
        bulk_composition = oil.sub_samples[0].bulk_composition
        for element in bulk_composition:
            if element.name == "vanadium":
                Va = element.measurement.value
    except Exception:
        Va = 0.

    try:
        f_asph = oil.sub_samples[0].SARA.asphaltenes.value
    except Exception:
        f_asph = 0.

    # need to go through subsamples checking for weathered data
    # if oil.metadata.emuls_constant_max is not None:
    # bullwinkle_fraction = emuls_constant_max

    if (oil.metadata.product_type != "Crude Oil NOS" and
            oil.metadata.product_type != "Bitumen Blend"):
        bullwinkle_fraction = 1.0
    else:
        oil_api = oil.metadata.API
        if (Ni > 0.0 and Va > 0.0 and Ni + Va > 15.0):
            bullwinkle_fraction = 0.0
        elif f_asph > 0:
            bullwinkle_fraction = 0.20219 - 0.168 * np.log10(f_asph)
            bullwinkle_fraction = np.clip(bullwinkle_fraction, 0.0, 0.303)
        elif oil_api < 26.0:
            bullwinkle_fraction = 0.08
        elif oil_api > 50.0:
            bullwinkle_fraction = 0.303
        else:
            bullwinkle_fraction = (-1.038 - 0.78935 * np.log10(1.0 / oil_api))

        bullwinkle_fraction = _adios2_new_bull_calc(bullwinkle_fraction,
                                                    oil_api)

    return bullwinkle_fraction


def _adios2_new_bull_calc(bullwinkle_fraction, oil_api):
    """
    From the Adios2 c++ file OilInitialize.cpp, there is functionality
    inside the function CAdiosData::Bullwinkle() which is annotated
    in the code as 'new bull calc'.

    It uses the following definitions:
    - TG, Documented as the value 'dT/df - evaporation'.
          I can only assume this is the initial fractional rate of
          evaporation.
    - TBP, Documented as the 'ADIOS 1 liquid boiling point
          (bubble pt)'.
    - BullAdios1, which appears to be used to scale-average the
                  initially computed bullwinkle fraction.
      get_density_data
    Regardless, in order to approximate what Adios2 is doing, we
    need this modification of our bullwinkle fraction.
    """

    t_g = 1356.7 - 247.36 * np.log(oil_api)
    t_bp = 532.98 - 3.1295 * oil_api
    bull_adios1 = (483.0 - t_bp) / t_g

    bull_adios1 = np.clip(bull_adios1, 0.0, 0.4)

    return 0.5 * (bullwinkle_fraction + bull_adios1)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['opendrift',
 'opendrift.elements',
 'opendrift.export',
 'opendrift.models',
 'opendrift.models.eulerdrift',
 'opendrift.models.openoil',
 'opendrift.models.openoil.adios',
 'opendrift.models.openoil.adios.computation',
 'opendrift.models.openoil.adios.extra_oils',
 'opendrift.models.openoil.adios.extra_oils.templates',
 'opendrift.models.openoil.adios.models',
 'opendrift.models.openoil.adios.models.common',
 'opendrift.models.openoil.adios.models.oil',
 'opendrift.models.openoil.adios.models.oil.cleanup',
 'opendrift.models.openoil.adios.models.oil.validation',
 'opendrift.models.openoil.adios.util',
 'opendrift.readers',
 'opendrift.readers.basereader',
 'opendrift.readers.interpolation',
 'opendrift.readers.operators',
 'opendrift.readers.roppy',
 'opendrift.readers.unstructured',
 'opendrift.scripts']

package_data = \
{'': ['*']}

install_requires = \
['Cartopy>=0.20',
 'cmocean>=2.0,<3.0',
 'coloredlogs>=15.0.1,<16.0.0',
 'geojson>=2.5.0,<3.0.0',
 'matplotlib>=3.5',
 'nc-time-axis>=1.4.1,<2.0.0',
 'netCDF4>=1.6',
 'numpy>=1.23',
 'pykdtree>=1.3.5,<2.0.0',
 'pynucos>=3',
 'pyproj>=2.3',
 'requests>=2.28.1,<3.0.0',
 'roaring-landmask>=0.5.1',
 'scipy>=1.9',
 'utm>=0.7.0,<0.8.0',
 'xarray>=2022.6.0']

extras_require = \
{'grib': ['cfgrib>=0.9.10,<0.10.0', 'pygrib>=2.1.4,<3.0.0']}

entry_points = \
{'console_scripts': ['hodograph = opendrift.scripts.hodograph:main',
                     'opendrift_animate = '
                     'opendrift.scripts.opendrift_animate:main',
                     'opendrift_gui = opendrift.scripts.opendrift_gui:main',
                     'opendrift_plot = opendrift.scripts.opendrift_plot:main',
                     'readerinfo = opendrift.scripts.readerinfo:main']}

setup_kwargs = {
    'name': 'opendrift',
    'version': '1.9.1rc2',
    'description': 'OpenDrift - a framework for ocean trajectory modeling',
    'long_description': 'None',
    'author': 'Knut-Frode Dagestad',
    'author_email': 'knutfd@met.no',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import numpy as np
from scipy.ndimage import map_coordinates
import scipy.ndimage as ndimage
from scipy.interpolate import interp1d, LinearNDInterpolator

from .interpolators import Nearest2DInterpolator, fill_NaN_towards_seafloor, horizontal_interpolation_methods, vertical_interpolation_methods
from opendrift.readers.basereader import variables

import logging
logger = logging.getLogger(__name__)

class ReaderBlock():
    """Class to store and interpolate the output from a reader with data on a regular (structured) grid."""

    def __init__(self, data_dict,
                 interpolation_horizontal='linearNDFast',
                 interpolation_vertical='linear'):

        # Make pointers to data values, for convenience
        self.x = data_dict['x']
        self.y = data_dict['y']
        self.time = data_dict['time']
        self.data_dict = data_dict
        del self.data_dict['x']
        del self.data_dict['y']
        del self.data_dict['time']
        try:
            self.z = data_dict['z']
            del self.data_dict['z']
        except:
            self.z = None

        # Mask any extremely large values, e.g. if missing netCDF _Fill_value
        filled_variables = set()
        for var in self.data_dict:
            self.data_dict[var] = variables.Variables.__check_variable_array__(var, self.data_dict[var])

            if isinstance(self.data_dict[var], np.ma.core.MaskedArray):
                self.data_dict[var] = np.ma.masked_outside(
                    np.ma.masked_invalid(self.data_dict[var]), -1E+9, 1E+9)
                # Convert masked arrays to numpy arrays
                self.data_dict[var] = np.ma.filled(self.data_dict[var],
                                                   fill_value=np.nan)
            # Fill missing data towards seafloor if 3D
            if isinstance(self.data_dict[var], (list,)):
                logger.warning('Ensemble data currently not extrapolated towards seafloor')
            elif self.data_dict[var].ndim == 3:
                filled = fill_NaN_towards_seafloor(self.data_dict[var])
                if filled is True:
                    filled_variables.add(var)

        if len(filled_variables) > 0:
            logger.debug('Filled NaN-values toward seafloor for :'
                          + str(list(filled_variables)))

        # Set 1D (vertical) and 2D (horizontal) interpolators
        try:
            self.Interpolator2DClass = \
                horizontal_interpolation_methods[interpolation_horizontal]
        except Exception:
            raise NotImplementedError(
                'Valid interpolation methods are: ' +
                str(horizontal_interpolation_methods.keys()))

        try:
            self.Interpolator1DClass = \
                vertical_interpolation_methods[interpolation_vertical]
        except Exception:
            raise NotImplementedError(
                'Valid interpolation methods are: ' +
                str(vertical_interpolation_methods.keys()))

        if 'land_binary_mask' in self.data_dict.keys() and \
                interpolation_horizontal != 'nearest':
            logger.debug('Nearest interpolation will be used '
                          'for landmask, and %s for other variables'
                          % interpolation_horizontal)

    def _initialize_interpolator(self, x, y, z=None):
        logger.debug('Initialising interpolator.')
        self.interpolator2d = self.Interpolator2DClass(self.x, self.y, x, y)
        if self.z is not None and len(np.atleast_1d(self.z)) > 1:
            self.interpolator1d = self.Interpolator1DClass(self.z, z)

    def interpolate(self, x, y, z=None, variables=None,
                    profiles=[], profiles_depth=None):

        self._initialize_interpolator(x, y, z)

        env_dict = {}
        if profiles is not []:
            profiles_dict = {'z': self.z}
        for varname, data in self.data_dict.items():
            nearest = False
            if varname == 'land_binary_mask':
                nearest = True
                self.interpolator2d_nearest = Nearest2DInterpolator(self.x, self.y, x, y)
            if type(data) is list:
                num_ensembles = len(data)
                logger.debug('Interpolating %i ensembles for %s' % (num_ensembles, varname))
                if data[0].ndim == 2:
                    horizontal = np.zeros(x.shape)*np.nan
                else:
                    horizontal = np.zeros((len(self.z), len(x)))*np.nan
                ensemble_number = np.remainder(range(len(x)), num_ensembles)
                for en in range(num_ensembles):
                    elnum = ensemble_number == en
                    int_full = self._interpolate_horizontal_layers(data[en], nearest=nearest)
                    if int_full.ndim == 1:
                        horizontal[elnum] = int_full[elnum]
                    else:
                        horizontal[:, elnum] = int_full[:, elnum]
            else:
                horizontal = self._interpolate_horizontal_layers(data, nearest=nearest)
            if profiles is not None and varname in profiles:
                profiles_dict[varname] = horizontal
            if horizontal.ndim > 1:
                env_dict[varname] = self.interpolator1d(horizontal)
            else:
                env_dict[varname] = horizontal
        if 'z' in profiles_dict:
            profiles_dict['z'] = np.atleast_1d(profiles_dict['z'])

        return env_dict, profiles_dict

    def _interpolate_horizontal_layers(self, data, nearest=False):
        '''Interpolate all layers of 3d (or 2d) array.'''

        if nearest is True:
            interpolator2d = self.interpolator2d_nearest
        else:
            interpolator2d = self.interpolator2d
        if data.ndim == 2:
            return interpolator2d(data)
        if data.ndim == 3:
            num_layers = data.shape[0]
            # Allocate output array
            result = np.ma.empty((num_layers, len(interpolator2d.x)))
            for layer in range(num_layers):
                result[layer, :] = self.interpolator2d(data[layer, :, :])
            return result

    def covers_positions(self, x, y, z=None):
        '''Check if given positions are covered by this reader block.'''

        indices = np.where((x >= self.x.min()) & (x <= self.x.max()) &
                           (y >= self.y.min()) & (y <= self.y.max()))[0]
        if len(indices) == len(x):
            return True
        else:
            return False




class fakeproj:
    # For readers with unprojected domain, we emulate a
    # pyproj class with needed functions
    class _crs:
        is_geographic = False

    crs = _crs()

    def __call__(self, x, y, inverse=False):
        # Simply return x and y since these are also row/column indices
        return x, y

    srs = 'None'





# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2020, Knut-Frode Dagestad, MET Norway

"""
SedimentDrift is an OpenDrift module for drift and settling of sediments.
Based on work by Simon Weppe, MetOcean Solutions Ltd.
"""

import numpy as np
import logging; logger = logging.getLogger(__name__)
from opendrift.models.oceandrift import OceanDrift
from opendrift.models.oceandrift import Lagrangian3DArray

class SedimentElement(Lagrangian3DArray):
    variables = Lagrangian3DArray.add_variables([
        ('settled', {'dtype': np.uint8,  # 0 is active, 1 is settled
                     'units': '1',
                     'default': 0}),
        ('terminal_velocity', {'dtype': np.float32,
                               'units': 'm/s',
                               'default': -0.001})  # 1 mm/s negative buoyancy
        ])


class SedimentDrift(OceanDrift):
    """Model for sediment drift, under development
    """

    ElementType = SedimentElement

    required_variables = {
        'x_sea_water_velocity': {'fallback': 0},
        'y_sea_water_velocity': {'fallback': 0},
        'upward_sea_water_velocity': {'fallback': 0},
        'x_wind': {'fallback': 0},
        'y_wind': {'fallback': 0},
        'sea_surface_wave_stokes_drift_x_velocity': {'fallback': 0},
        'sea_surface_wave_stokes_drift_y_velocity': {'fallback': 0},
        'sea_surface_wave_period_at_variance_spectral_density_maximum': {'fallback': 0},
        'sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment': {'fallback': 0},
        'land_binary_mask': {'fallback': None},
        'ocean_vertical_diffusivity': {'fallback': 0.02},
        'ocean_mixed_layer_thickness': {'fallback': 50},
        'sea_floor_depth_below_sea_level': {'fallback': 0},
        }

    def __init__(self, *args, **kwargs):
        """ Constructor of SedimentDrift module
        """

        super(SedimentDrift, self).__init__(*args, **kwargs)

        # By default, sediments do not strand towards coastline
        # TODO: A more sophisticated stranding algorithm is needed
        self._set_config_default('general:coastline_action', 'previous')

        # Vertical mixing is enabled as default
        self._set_config_default('drift:vertical_mixing', True)

    def update(self):
        """Update positions and properties of sediment particles.
        """

        # Advecting here all elements, but want to soon add
        # possibility of not moving settled elements, until
        # they are resuspended. May then need to send a boolean
        # array to advection methods below
        self.advect_ocean_current()

        self.vertical_advection()

        self.advect_wind()  # Wind shear in upper 10cm of ocean

        self.stokes_drift()

        self.vertical_mixing()  # Including buoyancy and settling

        self.resuspension()

    def bottom_interaction(self, seafloor_depth):
        """Sub method of vertical_mixing, determines settling"""
        # Elements at or below seafloor are settled, by setting
        # self.elements.moving to 0.
        # These elements will not move until eventual later resuspension.
        settling = np.logical_and(self.elements.z <= seafloor_depth, self.elements.moving==1)
        if np.sum(settling) > 0:
            logger.debug('Settling %s elements at seafloor' % np.sum(settling))
            self.elements.moving[settling] = 0

    def resuspension(self):
        """Resuspending elements if current speed > .5 m/s"""
        resuspending = np.logical_and(self.current_speed()>.5, self.elements.moving==0)
        if np.sum(resuspending) > 0:
            # Allow moving again
            self.elements.moving[resuspending] = 1
            # Suspend 1 cm above seafloor
            self.elements.z[resuspending] = self.elements.z[resuspending] + .01



from pathlib import Path
import os
import sys
import requests
from opendrift.models.openoil.adios.oil import OpendriftOil

import coloredlogs
import logging
logger = logging.getLogger('harvest_oils')

ADIOS = "https://adios.orr.noaa.gov/api/oils/"

def get_full_oil_from_id(_id):
    logger.debug(f"Fetching full oil: {_id}")
    o = requests.get(f"{ADIOS}/{_id}")
    o.raise_for_status()
    return OpendriftOil(o.json())

def oils(limit=9999, query=''):
    """
    Get all oils.

    Args:

        limit: number of oils to retrieve, None means all available.
        query: search string (name, id, location).


    Returns:

        List of `class:ThinOil`s.
    """

    o = requests.get(ADIOS, {
        'dir': 'asc',
        'limit': limit,
        'page': 0,
        'sort': 'metadata.name',
        'q': query
    })
    o.raise_for_status()
    o = o.json()

    oils = o['data']

    logging.debug(f"Total oils: {o['meta']['total']}, download limit: {limit}")
    logger.info(f"Fetched {len(oils)} oils from ADIOS")

    limit = len(oils) if limit is None else limit
    oils = oils[:min(limit, len(oils))]
    oils = [o['_id'] for o in oils]

    return oils

def download(dst):
    logger.info('downloading all oils..')

    ols = oils()
    logger.info(f'downloaded list of oils: {len(ols)}, fetching full oil..')

    for o in ols:
        logger.info(f"fetching oil: {o}..")
        o = get_full_oil_from_id(o)
        if o.valid():
            f = dst / Path(o.id).with_suffix('.json')
            with open(f, 'w') as fd:
                fd.write(o.json())
        else:
            logger.warning(f'skipping invalid oil: {o.id}')

def make_archive(oildir, file):
    logger.info(f'making archive: {oildir / "*.json"}')

    # Making a big JSON array with oils as a dictionary each.
    oils = []

    import glob
    import json
    import lzma
    for f in glob.glob(str(oildir / '*.json')):
        with open(f, 'r') as fd:
            oils.append(json.load(fd))

    logger.info(f'added {len(oils)} oils..')
    logger.info(f'compressing to {file}..')
    with lzma.open(file, 'wt') as c:
        json.dump(oils, c)


if __name__ == '__main__':
    coloredlogs.install('debug')

    dst = Path('oils')

    if not dst.exists():
        os.makedirs(dst)

    if not '--skip-download' in sys.argv:
        download(dst)

    make_archive(dst, 'oils.xz')




# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2020, Gaute Hope, MET Norway

import logging
logger = logging.getLogger(__name__)

import numpy as np
import xarray as xr
from datetime import datetime

from .basereader import BaseReader
from .basereader.structured import StructuredReader


class Reader(BaseReader, StructuredReader):
    dataset = None
    variables = None

    def __init__(self, filename, proj4=None, engine='cfgrib'):
        """
        Grib file reader

        Args:

            filename: path to grib or grib2 file.

            proj4: Optional projection override.

            engine: grib engine used by xarray, default is `cfgrib`.

        Returns: Grib-file reader.

        """
        logger.warning("This reader is experimental and may change in breaking ways without a major version change")

        self.dataset = xr.open_dataset(filename, engine=engine)
        self.name = filename

        self.times = np.atleast_1d(self.dataset.time.values)

        # convert to python datetime
        assert '[ns]' in self.times.dtype.str, "datetime is not in nanoseconds"
        self.times = np.array([
            datetime.utcfromtimestamp(d.astype('int') / 1e9)
            for d in self.times
        ])

        self.start_time = self.times[0]
        self.end_time = self.times[-1]
        if len(self.times) > 1:
            self.time_step = self.times[1] - self.times[0]
            assert np.all(
                np.diff(self.times) == self.time_step), "time step not constant"

        # scan for projection
        self.proj4 = proj4
        if self.proj4 is None:
            for v in self.dataset.variables:
                v = self.dataset[v]
                if 'GRIB_gridType' in v.attrs:
                    if self.proj4 is None:
                        self.proj4 = v.attrs['GRIB_gridType']
                        logger.info("projection: %s" % self.proj4)
                    else:
                        assert self.proj4 == v.attrs[
                            'GRIB_gridType'], "all variables must have the same projection"
        else:
            logger.info("Using supplied projection: %s" % self.proj4)

        # parsing variables
        self.variable_mapping = {}

        self.lon = self.dataset.longitude.values
        self.lat = self.dataset.latitude.values

        for v in self.dataset.variables:
            var = self.dataset[v]
            std_name = var.standard_name

            if std_name is not None:
                std_name = self.variable_aliases.get(std_name, std_name)
                self.variable_mapping[std_name] = v
                logger.debug("Found standard variable: %s" % v)

        self.variables = list(self.variable_mapping.keys())

        super().__init__()

        self._make_projected_grid_(self.lon, self.lat)

        self.set_buffer_size(5.)

    def get_variables(self,
                      requested_variables,
                      time=None,
                      x=None,
                      y=None,
                      z=None):

        requested_variables, time, x, y, z, _outside = self.check_arguments(
            requested_variables, time, x, y, z)

        nearestTime, dummy1, dummy2, indxTime, dummy3, dummy4 = \
            self.nearest_time(time)

        logger.debug("Request variables: %s" % requested_variables)

        ix0, ix1, iy0, iy1 = self._bbox_(x, y)

        variables = {}

        # This works because we have already asserted that delta_x and delta_y
        # are approx. constant.
        variables['y'] = self.x[ix0:ix1]
        variables['x'] = self.y[iy0:iy1]

        variables['time'] = nearestTime
        variables['z'] = z

        for v in requested_variables:
            par = v
            if hasattr(self, 'rotate_mapping') and par in self.rotate_mapping:
                logger.debug('Using %s to retrieve %s' %
                    (self.rotate_mapping[par], par))
                if par not in self.variable_mapping:
                    self.variable_mapping[par] = \
                        self.variable_mapping[
                            self.rotate_mapping[par]]

            var = self.variable_mapping[v]
            var = self.dataset[var]
            logger.debug("Fetching %s [%d:%d, %d:%d]" %
                         (v, ix0, ix1, iy0, iy1))
            variables[v] = self._slice_variable_(var, indxTime, slice(iy0, iy1), slice(ix0, ix1), 0, 0).values

        return variables

